# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""security management and error screens"""


import sys
import traceback

from logilab.common.registry import yes
from logilab.mtconverter import xml_escape

from cubicweb import _
from cubicweb.predicates import none_rset, match_user_groups, authenticated_user
from cubicweb.uilib import html_traceback, rest_traceback, exc_message
from cubicweb_web.view import AnyRsetView, EntityView, View
from cubicweb_web import formwidgets as wdgs
from cubicweb_web.formfields import guess_field
from cubicweb_web.views.schema import SecurityViewMixIn

SUBMIT_MSGID = _("Submit bug report")
MAIL_SUBMIT_MSGID = _("Submit bug report by mail")


class SecurityManagementView(SecurityViewMixIn, EntityView):
    """display security information for a given entity"""

    __regid__ = "security"
    __select__ = EntityView.__select__ & authenticated_user()

    title = _("security")

    def call(self):
        self.w('<div id="progress">%s</div>', self._cw._("validating..."))
        super(SecurityManagementView, self).call()

    def entity_call(self, entity):
        self._cw.add_js("cubicweb.edition.js")
        self._cw.add_css("cubicweb.acl.css")
        _ = self._cw._
        self.w(
            '<h1><span class="etype">%s</span> <a href="%s">%s</a></h1>',
            entity.dc_type().capitalize(),
            entity.absolute_url(),
            entity.dc_title(),
        )
        # first show permissions defined by the schema
        self.w("<h2>%s</h2>", _("Schema's permissions definitions"))
        self.permissions_table(entity.e_schema)
        self.w("<h2>%s</h2>", _("Manage security"))
        # ownership information
        if self._cw.vreg.schema.relation_schema_for("owned_by").has_perm(
            self._cw, "add", fromeid=entity.eid
        ):
            self.owned_by_edit_form(entity)
        else:
            self.owned_by_information(entity)

    def owned_by_edit_form(self, entity):
        self.w("<h3>%s</h3>", self._cw._("Ownership"))
        msg = self._cw._("ownerships have been changed")
        form = self._cw.vreg["forms"].select(
            "base",
            self._cw,
            entity=entity,
            form_renderer_id="onerowtable",
            submitmsg=msg,
            form_buttons=[wdgs.SubmitButton()],
            domid="ownership%s" % entity.eid,
            __redirectvid="security",
            __redirectpath=entity.rest_path(),
        )
        field = guess_field(
            entity.e_schema, self._cw.vreg.schema["owned_by"], req=self._cw
        )
        form.append_field(field)
        form.render(w=self.w, display_progress_div=False)

    def owned_by_information(self, entity):
        ownersrset = entity.related("owned_by")
        if ownersrset:
            self.w("<h3>%s</h3>", self._cw._("Ownership"))
            self.w('<div class="ownerInfo">')
            self.w(self._cw._("this entity is currently owned by"), " ")
            self.wview("csv", entity.related("owned_by"), "null")
            self.w("</div>")
        # else we don't know if this is because entity has no owner or becayse
        # user as no access to owner users entities


class ErrorView(AnyRsetView):
    """default view when no result has been found"""

    __select__ = yes()
    __regid__ = "error"

    def page_title(self):
        """returns a title according to the result set - used for the
        title in the HTML header
        """
        return self._cw._("an error occurred")

    def _excinfo(self):
        req = self._cw
        ex = req.data.get("ex")
        excinfo = req.data.get("excinfo")
        if "errmsg" in req.data:
            errmsg = req.data["errmsg"]
            exclass = None
        else:
            errmsg = exc_message(ex, req.encoding)
            exclass = ex.__class__.__name__
        return errmsg, exclass, excinfo

    def call(self):
        req = self._cw.reset_headers()
        title = self._cw._("an error occurred")
        self.w("<h2>%s</h2>", title)

        ex, exclass, excinfo = self._excinfo()
        # save excinfo on req for easier debugging in interactive shell
        req.excinfo = ex, exclass, excinfo

        # if excinfo is not None, it's probably not a bug
        if excinfo is None:
            return

        # only render traceback in debugmode
        if self._cw.vreg.config.debugmode:
            if self._cw.vreg.config["print-traceback"]:
                if exclass is None:
                    self.w(
                        '<div class="tb">%s</div>',
                        xml_escape(ex).replace("\n", "<br />"),
                        escape=False,
                    )
                else:
                    self.w(
                        '<div class="tb">%s: %s</div>',
                        xml_escape(exclass),
                        xml_escape(ex).replace("\n", "<br />"),
                        escape=False,
                    )
                self.w("<hr />")
                self.w(
                    '<div class="tb">%s</div>',
                    html_traceback(excinfo, ex, ""),
                    escape=False,
                )
            else:
                self.w(
                    '<div class="tb">%s</div>',
                    xml_escape(ex).replace("\n", "<br />"),
                    escape=False,
                )
        else:
            self.w(
                "<div>Error 500, please contact the server administrator for more information</div>"
            )

        # displaying the exception of stderr for debugging
        traceback.print_tb(tb=excinfo[-1])
        print()
        print(excinfo[1], file=sys.stderr)

        if self._cw.cnx:
            vcconf = self._cw.cnx.repo.get_versions()
            self.w("<div>")
            eversion = vcconf.get("cubicweb", self._cw._("no version information"))
            # NOTE: tuple wrapping needed since eversion is itself a tuple
            self.w("<b>CubicWeb version:</b> %s<br/>\n", eversion)
            cversions = []
            for cube in self._cw.vreg.config.cubes():
                cubeversion = vcconf.get(cube, self._cw._("no version information"))
                self.w("<b>Cube %s version:</b> %s<br/>\n", cube, cubeversion)
                cversions.append((cube, cubeversion))
            self.w("</div>")
        else:
            eversion = self._cw._("no version information")
            cversions = []
        # creates a bug submission link if submit-mail is set
        if self._cw.vreg.config["submit-mail"]:
            form = self._cw.vreg["forms"].select(
                "base", self._cw, rset=None, mainform=False
            )
            binfo = text_error_description(ex, excinfo, req, eversion, cversions)
            form.add_hidden(
                "description",
                binfo,
                # we must use a text area to keep line breaks
                widget=wdgs.TextArea({"class": "hidden"}),
            )
            # add a signature so one can't send arbitrary text
            form.add_hidden("__signature", req.vreg.config.sign_text(binfo))
            form.add_hidden("__bugreporting", "1")
            form.form_buttons = [wdgs.SubmitButton(MAIL_SUBMIT_MSGID)]
            form.action = req.build_url("reportbug")
            form.render(w=self.w)


def text_error_description(ex, excinfo, req, eversion, cubes):
    binfo = rest_traceback(excinfo, xml_escape(ex))
    binfo += "\n\n:URL: %s\n" % req.url()
    if "__bugreporting" not in req.form:
        binfo += "\n:form params:\n"
        binfo += "\n".join("  * %s = %s" % (k, v) for k, v in req.form.items())
    binfo += "\n\n:CubicWeb version: %s\n" % (eversion,)
    for pkg, pkgversion in cubes:
        binfo += ":Cube %s version: %s\n" % (pkg, pkgversion)
    binfo += "\n"
    return binfo


class CwStats(View):
    """A textual stats output for monitoring tools such as munin"""

    __regid__ = "processinfo"
    content_type = "text/plain"
    templatable = False
    __select__ = none_rset() & match_user_groups("users", "managers")

    def call(self):
        stats = self._cw.call_service("repo_stats")
        stats["threads"] = ", ".join(sorted(stats["threads"]))
        for k in stats:
            if k in ("extid_cache_size", "type_source_cache_size"):
                continue
            if k.endswith("_cache_size"):
                stats[k] = "%s / %s" % (stats[k]["size"], stats[k]["maxsize"])
        results = []
        for element in stats:
            results.append("%s %s" % (element, stats[element]))
        self.w("\n".join(results))



# copyright 2013-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.

"""cubicweb-forum views/forms/actions/components for web ui"""

from cubicweb.predicates import is_instance

from cubicweb_web import view
from cubicweb_web.views import primary, baseviews, uicfg
from cubicweb_web.views.uicfg import autoform_section as afs


class MyAFS(uicfg.AutoformSectionRelationTags):
    __select__ = is_instance("ForumThread")


_myafs = MyAFS(__module__=__name__)

_myafs.tag_object_of(("*", "in_forum", "Forum"), "main", "inlined")

afs.tag_object_of(("*", "in_forum", "Forum"), "main", "inlined")


class ForumSameETypeListView(baseviews.SameETypeListView):
    __select__ = baseviews.SameETypeListView.__select__ & is_instance("Forum")

    def call(self, **kwargs):
        _ = self._cw._
        _("Topic"), _("Description")
        _("Number of threads"), _("Last activity")
        _(
            """a long
tranlated line
hop."""
        )


class ForumLastActivity(view.EntityView):
    __regid__ = "forum_last_activity"
    __select__ = view.EntityView.__select__ & is_instance("Forum")


class ForumPrimaryView(primary.PrimaryView):
    __select__ = primary.PrimaryView.__select__ & is_instance("Forum")

    def render_entity_attributes(self, entity):
        _ = self._cw._
        _("Subject"), _("Created"), _("Answers"),
        _("Last answered")
        _("This forum does not have any thread yet.")


class ForumThreadPrimaryView(primary.PrimaryView):
    __select__ = primary.PrimaryView.__select__ & is_instance("ForumThread")



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Specific views for schema related entities"""


import codecs
import os
import tempfile
from itertools import cycle

from logilab.common.graph import GraphGenerator, DotBackend
from logilab.common.registry import yes
from logilab.mtconverter import xml_escape
from yams import BASE_TYPES, schema2dot as s2d

from cubicweb import _
from cubicweb import uilib
from cubicweb.predicates import (
    is_instance,
    has_related_entities,
    authenticated_user,
)
from cubicweb.schema import (
    META_RTYPES,
    SCHEMA_TYPES,
    SYSTEM_RTYPES,
    WORKFLOW_TYPES,
    INTERNAL_TYPES,
)
from cubicweb_web.view import EntityView, StartupView
from cubicweb_web import action, facet, schemaviewer
from cubicweb_web.views import uicfg, baseviews, tabs, tableview, ibreadcrumbs

ALWAYS_SKIP_TYPES = BASE_TYPES | SCHEMA_TYPES
SKIP_TYPES = (
    ALWAYS_SKIP_TYPES | META_RTYPES | SYSTEM_RTYPES | WORKFLOW_TYPES | INTERNAL_TYPES
)
SKIP_TYPES.update({"CWUser", "CWGroup", "EmailAddress", "Bookmark"})


def skip_types(req):
    if int(req.form.get("skipmeta", True)):
        return SKIP_TYPES
    return ALWAYS_SKIP_TYPES


_pvs = uicfg.primaryview_section
_pvdc = uicfg.primaryview_display_ctrl

for _action in ("read", "add", "update", "delete"):
    _pvs.tag_subject_of(("*", "%s_permission" % _action, "*"), "hidden")
    _pvs.tag_object_of(("*", "%s_permission" % _action, "*"), "hidden")

for _etype in ("CWEType", "CWRType", "CWAttribute", "CWRelation"):
    _pvdc.tag_attribute((_etype, "description"), {"showlabel": False})

_pvs.tag_attribute(("CWEType", "name"), "hidden")
_pvs.tag_attribute(("CWEType", "final"), "hidden")
_pvs.tag_object_of(("*", "workflow_of", "CWEType"), "hidden")
_pvs.tag_subject_of(("CWEType", "default_workflow", "*"), "hidden")
_pvs.tag_object_of(("*", "specializes", "CWEType"), "hidden")
_pvs.tag_subject_of(("CWEType", "specializes", "*"), "hidden")
_pvs.tag_object_of(("*", "from_entity", "CWEType"), "hidden")
_pvs.tag_object_of(("*", "to_entity", "CWEType"), "hidden")

_pvs.tag_attribute(("CWRType", "name"), "hidden")
_pvs.tag_attribute(("CWRType", "final"), "hidden")
_pvs.tag_object_of(("*", "relation_type", "CWRType"), "hidden")

_pvs.tag_subject_of(("CWAttribute", "constrained_by", "*"), "hidden")
_pvs.tag_subject_of(("CWRelation", "constrained_by", "*"), "hidden")


class SecurityViewMixIn:
    """mixin providing methods to display security information for a entity,
    relation or relation definition schema
    """

    cssclass = "listing schemaInfo"

    def permissions_table(self, erschema, permissions=None):
        self._cw.add_css("cubicweb.acl.css")
        _ = self._cw._
        self.w('<table class="%s">' % self.cssclass)
        self.w(
            "<tr><th>%s</th><th>%s</th><th>%s</th></tr>"
            % (_("permission"), _("granted to groups"), _("rql expressions"))
        )
        for an_action in erschema.ACTIONS:
            self.w("<tr><td>%s</td><td>", _(an_action))
            if permissions is None:
                groups = erschema.get_groups(an_action)
                rqlexprs = sorted(
                    e.expression for e in erschema.get_rqlexprs(an_action)
                )
            else:
                groups = permissions[an_action][0]
                rqlexprs = permissions[an_action][1]
            # XXX get group entity and call it's incontext view
            groups = [
                '<a class="%s" href="%s">%s</a>'
                % (group, self._cw.build_url("cwgroup/%s" % group), label)
                for label, group in sorted((_(g), g) for g in groups)
            ]
            self.w("<br/>".join(groups))
            self.w("</td><td>")
            self.w("<br/>".join(rqlexprs))
            self.w("</td></tr>\n")
        self.w("</table>")

    def grouped_permissions_table(self, rschema):
        # group relation definitions with identical permissions
        perms = {}
        for rdef in rschema.relation_definition.values():
            rdef_perms = []
            for an_action in rdef.ACTIONS:
                groups = sorted(rdef.get_groups(an_action))
                exprs = sorted(e.expression for e in rdef.get_rqlexprs(an_action))
                rdef_perms.append((an_action, (tuple(groups), tuple(exprs))))
            rdef_perms = tuple(rdef_perms)
            if rdef_perms in perms:
                perms[rdef_perms].append((rdef.subject, rdef.object))
            else:
                perms[rdef_perms] = [(rdef.subject, rdef.object)]
        # set layout permissions in a table for each group of relation
        # definition
        _ = self._cw._
        self.w('<div style="margin: 0px 1.5em">')
        tmpl = "<strong>%s</strong> %s <strong>%s</strong>"
        for perm, rdefs in perms.items():
            self.w(
                "<div>%s</div>",
                ", ".join(
                    tmpl % (_(s.type), _(rschema.type), _(o.type)) for s, o in rdefs
                ),
            )
            # accessing rdef from previous loop by design: only used to get
            # ACTIONS
            self.permissions_table(rdef, dict(perm))
        self.w("</div>")


# global schema view ###########################################################


class SchemaView(tabs.TabsMixin, StartupView):
    """display schema information (graphically, listing tables...) in tabs"""

    __regid__ = "schema"
    title = _("data model schema")
    tabs = [_("schema-diagram"), _("schema-entity-types"), _("schema-relation-types")]
    default_tab = "schema-diagram"

    def call(self):
        self.w("<h1>%s</h1>", self._cw._(self.title))
        self.render_tabs(self.tabs, self.default_tab)


class SchemaImageTab(StartupView):
    __regid__ = "schema-diagram"

    def call(self):
        _ = self._cw._
        self.w(
            self._cw._(
                "<div>This schema of the data model <em>excludes</em> the "
                'meta-data, but you can also display a <a href="%s">complete '
                "schema with meta-data</a>.</div>"
            ),
            self._cw.build_url("view", vid="schemagraph", skipmeta=0),
        )
        self.w(
            '<div><a href="%s">%s</a></div>',
            self._cw.build_url("view", vid="owl"),
            self._cw._("Download schema as OWL"),
        )
        self.wview("schemagraph")


class SchemaETypeTab(StartupView):
    __regid__ = "schema-entity-types"

    def call(self):
        self.wview(
            "table",
            self._cw.execute(
                "Any X ORDERBY N WHERE X is CWEType, X name N, X final FALSE"
            ),
        )


class SchemaRTypeTab(StartupView):
    __regid__ = "schema-relation-types"

    def call(self):
        self.wview(
            "table",
            self._cw.execute(
                "Any X ORDERBY N WHERE X is CWRType, X name N, X final FALSE"
            ),
        )


# CWEType ######################################################################

# register msgid generated in entity relations tables
_("i18ncard_1"), _("i18ncard_?"), _("i18ncard_+"), _("i18ncard_*")


class CWETypePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("CWEType")
    tabs = [
        _("cwetype-description"),
        _("cwetype-box"),
        _("cwetype-workflow"),
        _("cwetype-views"),
        _("cwetype-permissions"),
    ]
    default_tab = "cwetype-description"


class CWETypeDescriptionTab(tabs.PrimaryTab):
    __regid__ = "cwetype-description"
    __select__ = tabs.PrimaryTab.__select__ & is_instance("CWEType")

    def render_entity_attributes(self, entity):
        super().render_entity_attributes(entity)
        _ = self._cw._
        # inheritance
        if entity.specializes:
            self.w("<div><strong>%s</strong>", _("Parent class:"))
            self.wview("csv", entity.related("specializes", "subject"))
            self.w("</div>")
        if entity.reverse_specializes:
            self.w("<div><strong>%s</strong>", _("Sub-classes:"))
            self.wview("csv", entity.related("specializes", "object"))
            self.w("</div>")
        # entity schema image
        self.wview("schemagraph", etype=entity.name)
        # entity schema attributes
        self.w("<h2>%s</h2>", _("CWAttribute_plural"))
        rset = self._cw.execute(
            "Any A,ON,D,C,A,DE,A, IDX,FTI,I18N,R,O,RN,S ORDERBY AA "
            "WHERE A is CWAttribute, A from_entity S, S eid %(x)s, "
            "A ordernum AA, A defaultval D, A description DE, A cardinality C, "
            "A fulltextindexed FTI, A internationalizable I18N, A indexed IDX, "
            "A relation_type R, R name RN, A to_entity O, O name ON",
            {"x": entity.eid},
        )
        self.wview(
            "table",
            rset,
            "null",
            cellvids={
                0: "rdef-name-cell",
                2: "etype-attr-defaultval-cell",
                3: "etype-attr-cardinality-cell",
                4: "rdef-constraints-cell",
                6: "rdef-options-cell",
            },
            headers=(
                _("name"),
                _("type"),
                _("default value"),
                _("required"),
                _("constraints"),
                _("description"),
                _("options"),
            ),
        )
        # entity schema relations
        self.w("<h2>%s</h2>", _("CWRelation_plural"))
        cellvids = {
            0: "rdef-name-cell",
            2: "etype-rel-cardinality-cell",
            3: "rdef-constraints-cell",
            4: "rdef-options-cell",
        }
        headers = [
            _("name"),
            _("object type"),
            _("cardinality"),
            _("constraints"),
            _("options"),
        ]
        rset = self._cw.execute(
            'Any A,TT,"i18ncard_"+SUBSTRING(C,1,1),A,A, K,TTN,R,RN ORDERBY RN '
            "WHERE A is CWRelation, A from_entity S, S eid %(x)s, "
            "A composite K, A cardinality C, "
            "A relation_type R, R name RN, A to_entity TT, TT name TTN",
            {"x": entity.eid},
        )
        if rset:
            self.w("<h5>%s %s</h5>", entity.name, _("is subject of:"))
            self.wview("table", rset, cellvids=cellvids, headers=headers)
        rset = self._cw.execute(
            'Any A,TT,"i18ncard_"+SUBSTRING(C,1,1),A,A, K,TTN,R,RN ORDERBY RN '
            "WHERE A is CWRelation, A to_entity O, O eid %(x)s, "
            "A composite K, A cardinality C, "
            "A relation_type R, R name RN, A from_entity TT, TT name TTN",
            {"x": entity.eid},
        )
        if rset:
            cellvids[0] = "rdef-object-name-cell"
            headers[1] = _("subject type")
            self.w("<h5>%s %s</h5>", entity.name, _("is object of:"))
            self.wview("table", rset, cellvids=cellvids, headers=headers)


class CWETypeAttributeCardinalityCell(baseviews.FinalView):
    __regid__ = "etype-attr-cardinality-cell"

    def cell_call(self, row, col):
        if self.cw_rset.rows[row][col][0] == "1":
            self.w(self._cw._("yes"))
        else:
            self.w(self._cw._("no"))


class CWETypeAttributeDefaultValCell(baseviews.FinalView):
    __regid__ = "etype-attr-defaultval-cell"

    def cell_call(self, row, col):
        defaultval = self.cw_rset.rows[row][col]
        if defaultval is not None:
            self.w(str(self.cw_rset.rows[row][col].unzpickle()))


class CWETypeRelationCardinalityCell(baseviews.FinalView):
    __regid__ = "etype-rel-cardinality-cell"

    def cell_call(self, row, col):
        self.w(self._cw._(self.cw_rset.rows[row][col]))


class CWETypeBoxTab(EntityView):
    __regid__ = "cwetype-box"
    __select__ = is_instance("CWEType")

    def cell_call(self, row, col):
        viewer = schemaviewer.SchemaViewer(self._cw)
        entity = self.cw_rset.get_entity(row, col)
        eschema = self._cw.vreg.schema.eschema(entity.name)
        layout = viewer.visit_entityschema(eschema)
        self.w(uilib.ureport_as_html(layout))
        self.w('<br class="clear"/>')


class CWETypePermTab(SecurityViewMixIn, EntityView):
    __regid__ = "cwetype-permissions"
    __select__ = is_instance("CWEType") & authenticated_user()

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        eschema = self._cw.vreg.schema.eschema(entity.name)
        self.w("<h4>%s</h4>", self._cw._("This entity type permissions:"))
        self.permissions_table(eschema)
        self.w('<div style="margin: 0px 1.5em">')
        self.w("<h4>%s</h4>", self._cw._("Attributes permissions:"))
        for attr, etype in eschema.attribute_definitions():
            if attr not in META_RTYPES:
                rdef = eschema.relation_definition(attr)
                attrtype = str(rdef.rtype)
                self.w(
                    '<h4 class="schema">%s (%s)</h4> ', attrtype, self._cw._(attrtype)
                )
                self.permissions_table(rdef)
        self.w("</div>")


class CWETypeWorkflowTab(EntityView):
    __regid__ = "cwetype-workflow"
    __select__ = is_instance("CWEType") & has_related_entities("workflow_of", "object")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        if entity.default_workflow:
            wf = entity.default_workflow[0]
            if len(entity.reverse_workflow_of) > 1:
                self.w("<h1>%s (%s)</h1>", wf.name, self._cw._("default_workflow"))
            self.display_workflow(wf)
            defaultwfeid = wf.eid
        else:
            self.w(
                '<div class="error">%s</div>',
                self._cw._("There is no default workflow"),
            )
            defaultwfeid = None
        for altwf in entity.reverse_workflow_of:
            if altwf.eid == defaultwfeid:
                continue
            self.w("<h1>%s</h1>", altwf.name)
            self.display_workflow(altwf)

    def display_workflow(self, wf):
        self.w(wf.view("wfgraph"))
        self.w(
            '<a href="%s">%s</a>',
            wf.absolute_url(),
            self._cw._("more info about this workflow"),
        )


class CWETypeViewsTab(EntityView):
    """possible views for this entity type"""

    __regid__ = "cwetype-views"
    __select__ = EntityView.__select__ & is_instance("CWEType")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        _ = self._cw._
        self.w(
            "<div>%s</div>",
            _(
                "Non exhaustive list of views that may "
                "apply to entities of this type"
            ),
        )
        views = [
            (view.content_type, view.__regid__, _(view.title))
            for view in self.possible_views(entity.name)
        ]
        self.wview(
            "pyvaltable",
            pyvalue=sorted(views),
            headers=(_("content type"), _("view identifier"), _("view title")),
        )

    def possible_views(self, etype):
        rset = self._cw.etype_rset(etype)
        return [
            v
            for v in self._cw.vreg["views"].possible_views(self._cw, rset)
            if v.category != "startupview"
        ]


class CWETypeOneLineView(baseviews.OneLineView):
    __select__ = is_instance("CWEType")

    def cell_call(self, row, col, **kwargs):
        entity = self.cw_rset.get_entity(row, col)
        if entity.final:
            self.w('<em class="finalentity">')
        super().cell_call(row, col, **kwargs)
        if entity.final:
            self.w("</em>")


# CWRType ######################################################################


class CWRTypePrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("CWRType")
    tabs = [_("cwrtype-description"), _("cwrtype-permissions")]
    default_tab = "cwrtype-description"


class CWRTypeDescriptionTab(tabs.PrimaryTab):
    __regid__ = "cwrtype-description"
    __select__ = is_instance("CWRType")

    def render_entity_attributes(self, entity):
        super().render_entity_attributes(entity)
        _ = self._cw._
        if not entity.final:
            self.wview("schemagraph", rtype=entity.name)
        rset = self._cw.execute(
            "Any R,C,R,R, RT WHERE "
            "R relation_type RT, RT eid %(x)s, "
            "R cardinality C",
            {"x": entity.eid},
        )
        self.wview(
            "table",
            rset,
            "null",
            headers=(_("relation"), _("cardinality"), _("constraints"), _("options")),
            cellvids={2: "rdef-constraints-cell", 3: "rdef-options-cell"},
        )


class CWRTypePermTab(SecurityViewMixIn, EntityView):
    __regid__ = "cwrtype-permissions"
    __select__ = is_instance("CWRType") & authenticated_user()

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rschema = self._cw.vreg.schema.rschema(entity.name)
        self.grouped_permissions_table(rschema)


# CWAttribute / CWRelation #####################################################


class RDEFPrimaryView(tabs.TabbedPrimaryView):
    __select__ = is_instance("CWRelation", "CWAttribute")
    tabs = [_("rdef-description"), _("rdef-permissions")]
    default_tab = "rdef-description"


class RDEFDescriptionTab(tabs.PrimaryTab):
    __regid__ = "rdef-description"
    __select__ = is_instance("CWRelation", "CWAttribute")

    def render_entity_attributes(self, entity):
        super().render_entity_attributes(entity)
        rdef = entity.yams_schema()
        if rdef.constraints:
            self.w("<h4>%s</h4>", self._cw._("constrained_by"))
            self.w(entity.view("rdef-constraints-cell"))


class RDEFPermTab(SecurityViewMixIn, EntityView):
    __regid__ = "rdef-permissions"
    __select__ = is_instance("CWRelation", "CWAttribute") & authenticated_user()

    def cell_call(self, row, col):
        self.permissions_table(self.cw_rset.get_entity(row, col).yams_schema())


class RDEFNameView(tableview.CellView):
    """display relation name and its translation only in a cell view, link to
    relation definition's primary view (for use in entity type relations table
    for instance)
    """

    __regid__ = "rdef-name-cell"
    __select__ = is_instance("CWRelation", "CWAttribute")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rtype = entity.relation_type[0].name
        # XXX use context entity + pgettext
        self.w(
            '<a href="%s">%s</a> (%s)', entity.absolute_url(), rtype, self._cw._(rtype)
        )


class RDEFObjectNameView(tableview.CellView):
    """same as RDEFNameView but when the context is the object entity"""

    __regid__ = "rdef-object-name-cell"
    __select__ = is_instance("CWRelation", "CWAttribute")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rtype = entity.relation_type[0].name
        # XXX use context entity + pgettext
        self.w(
            '<a href="%s">%s</a> (%s)',
            entity.absolute_url(),
            rtype,
            self._cw.__(rtype + "_object"),
        )


class RDEFConstraintsCell(EntityView):
    __regid__ = "rdef-constraints-cell"
    __select__ = is_instance("CWAttribute", "CWRelation")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rschema = self._cw.vreg.schema.rschema(entity.rtype.name)
        rdef = rschema.relation_definition(entity.stype.name, entity.otype.name)
        constraints = [xml_escape(str(c)) for c in getattr(rdef, "constraints")]
        self.w("<br/>".join(constraints))


class CWAttributeOptionsCell(EntityView):
    __regid__ = "rdef-options-cell"
    __select__ = is_instance("CWAttribute")

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        options = []
        if entity.indexed:
            options.append(self._cw._("indexed"))
        if entity.fulltextindexed:
            options.append(self._cw._("fulltextindexed"))
        if entity.internationalizable:
            options.append(self._cw._("internationalizable"))
        self.w(",".join(options))


class CWRelationOptionsCell(EntityView):
    __regid__ = "rdef-options-cell"
    __select__ = is_instance(
        "CWRelation",
    )

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        rtype = entity.rtype
        options = []
        if rtype.symmetric:
            options.append(self._cw._("symmetric"))
        if rtype.inlined:
            options.append(self._cw._("inlined"))
        if rtype.fulltext_container:
            options.append(
                "%s=%s"
                % (
                    self._cw._("fulltext_container"),
                    self._cw._(rtype.fulltext_container),
                )
            )
        if entity.composite:
            options.append(
                "{}={}".format(self._cw._("composite"), self._cw._(entity.composite))
            )
        self.w(",".join(options))


# schema images ###############################################################


class RestrictedSchemaVisitorMixIn:
    def __init__(self, req, *args, **kwargs):
        self._cw = req
        super().__init__(*args, **kwargs)

    def should_display_schema(self, rschema):
        return super().should_display_schema(rschema) and rschema.may_have_permission(
            "read", self._cw
        )

    def should_display_attr(self, eschema, rschema):
        return super().should_display_attr(
            eschema, rschema
        ) and eschema.relation_definition(rschema).may_have_permission("read", self._cw)


class FullSchemaVisitor(RestrictedSchemaVisitorMixIn, s2d.FullSchemaVisitor):
    pass


class OneHopESchemaVisitor(RestrictedSchemaVisitorMixIn, s2d.OneHopESchemaVisitor):
    pass


class OneHopRSchemaVisitor(RestrictedSchemaVisitorMixIn, s2d.OneHopRSchemaVisitor):
    pass


class CWSchemaDotPropsHandler(s2d.SchemaDotPropsHandler):
    def __init__(self, visitor, cw):
        self.visitor = visitor
        self.cw = cw
        self._cycle = iter(cycle(("#ff7700", "#000000", "#ebbc69", "#888888")))
        self.nextcolor = lambda: next(self._cycle)

        self.colors = {}

    def node_properties(self, eschema):
        """return DOT drawing options for an entity schema include href"""
        label = ["{", eschema.type, "|"]
        label.append(
            r"\l".join(
                "{} ({})".format(rel.type, eschema.relation_definition(rel.type).object)
                for rel in eschema.ordered_relations()
                if rel.final and self.visitor.should_display_attr(eschema, rel)
            )
        )
        label.append(r"\l}")  # trailing \l ensure alignement of the last one
        return {
            "label": "".join(label),
            "shape": "record",
            "fontname": "Courier",
            "style": "filled",
            "href": self.cw.build_url("cwetype/%s" % eschema.type),
            "fontsize": "10px",
        }

    def edge_properties(self, rschema, subjnode, objnode):
        """return default DOT drawing options for a relation schema"""
        # Inheritance relation (i.e 'specializes').
        if rschema is None:
            kwargs = {
                "label": "Parent class",
                "color": "grey",
                "style": "filled",
                "arrowhead": "empty",
                "fontsize": "10px",
            }
        # symmetric rels are handled differently, let yams decide what's best
        elif rschema.symmetric:
            kwargs = {
                "label": rschema.type,
                "color": "#887788",
                "style": "dashed",
                "dir": "both",
                "arrowhead": "normal",
                "arrowtail": "normal",
                "fontsize": "10px",
                "href": self.cw.build_url("cwrtype/%s" % rschema.type),
            }
        else:
            kwargs = {
                "label": rschema.type,
                "color": "black",
                "style": "filled",
                "fontsize": "10px",
                "href": self.cw.build_url("cwrtype/%s" % rschema.type),
            }
            rdef = rschema.relation_definition(subjnode, objnode)
            if rdef.composite == "subject":
                kwargs["arrowhead"] = "none"
                kwargs["arrowtail"] = "diamond"
            elif rdef.composite == "object":
                kwargs["arrowhead"] = "diamond"
                kwargs["arrowtail"] = "none"
            else:
                kwargs["arrowhead"] = "open"
                kwargs["arrowtail"] = "none"
            # UML like cardinalities notation, omitting 1..1
            if rdef.cardinality[1] != "1":
                kwargs["taillabel"] = s2d.CARD_MAP[rdef.cardinality[1]]
            if rdef.cardinality[0] != "1":
                kwargs["headlabel"] = s2d.CARD_MAP[rdef.cardinality[0]]
            try:
                kwargs["color"] = self.colors[rschema]
            except KeyError:
                kwargs["color"] = self.nextcolor()
                self.colors[rschema] = kwargs["color"]
        kwargs["fontcolor"] = kwargs["color"]
        # dot label decoration is just awful (1 line underlining the label
        # + 1 line going to the closest edge spline point)
        kwargs["decorate"] = "false"
        # kwargs['labelfloat'] = 'true'
        return kwargs


class SchemaGraphView(StartupView):
    __regid__ = "schemagraph"

    def call(self, etype=None, rtype=None, alt=""):
        schema = self._cw.vreg.schema
        if etype:
            assert rtype is None
            visitor = OneHopESchemaVisitor(
                self._cw, schema.eschema(etype), skiptypes=skip_types(self._cw)
            )
            alt = self._cw._(
                "graphical representation of the %(etype)s "
                "entity type from %(appid)s data model"
            )
        elif rtype:
            visitor = OneHopRSchemaVisitor(
                self._cw, schema.rschema(rtype), skiptypes=skip_types(self._cw)
            )
            alt = self._cw._(
                "graphical representation of the %(rtype)s "
                "relation type from %(appid)s data model"
            )
        else:
            visitor = FullSchemaVisitor(
                self._cw, schema, skiptypes=skip_types(self._cw)
            )
            alt = self._cw._("graphical representation of %(appid)s data model")
        alt %= {"rtype": rtype, "etype": etype, "appid": self._cw.vreg.config.appid}
        prophdlr = CWSchemaDotPropsHandler(visitor, self._cw)
        generator = GraphGenerator(
            DotBackend(
                "schema",
                "BT",
                ratio="compress",
                size=None,
                renderer="dot",
                additionnal_param={
                    "overlap": "false",
                    "splines": "true",
                    "sep": "0.2",
                },
            )
        )
        # svg image file
        fd, tmpfile = tempfile.mkstemp(".svg")
        try:
            os.close(fd)
            generator.generate(visitor, prophdlr, tmpfile)
            with codecs.open(tmpfile, "rb", encoding="utf-8") as svgfile:
                self.w(svgfile.read())
        finally:
            os.unlink(tmpfile)


# breadcrumbs ##################################################################


class CWRelationIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("CWRelation")

    def parent_entity(self):
        return self.entity.rtype


class CWAttributeIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("CWAttribute")

    def parent_entity(self):
        return self.entity.stype


class CWConstraintIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("CWConstraint")

    def parent_entity(self):
        if self.entity.reverse_constrained_by:
            return self.entity.reverse_constrained_by[0]


class RQLExpressionIBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    __select__ = is_instance("RQLExpression")

    def parent_entity(self):
        return self.entity.expression_of


# misc: facets, actions ########################################################


class CWFinalFacet(facet.AttributeFacet):
    __regid__ = "cwfinal-facet"
    __select__ = facet.AttributeFacet.__select__ & is_instance("CWEType", "CWRType")
    rtype = "final"


class ViewSchemaAction(action.Action):
    __regid__ = "schema"
    __select__ = yes()

    title = _("data model schema")
    order = 30
    category = "manage"

    def url(self):
        return self._cw.build_url(self.__regid__)



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common.testlib import TestCase, unittest_main

from cubicweb_web import Redirect
from cubicweb_web.views.apacherewrite import ApacheURLRewrite, RewriteCond


class ApacheURLRewriteTC(TestCase):
    def test(self):
        class MyAppRules(ApacheURLRewrite):
            rules = [
                RewriteCond(
                    r"logilab\.fr",
                    match="host",
                    rules=[("/(.*)", r"http://www.logilab.fr/\1")],
                    action="redirect",
                ),
                RewriteCond(r"(www)\.logilab\.fr", match="host", action="stop"),
                RewriteCond("/(data|json)/", match="path", action="stop"),
                RewriteCond(
                    r"(?P<cat>.*)\.logilab\.fr",
                    match="host",
                    rules=[("/(.*)", r"/m_%(cat)s/\1")],
                ),
            ]

        urlrewriter = MyAppRules()
        req = None  # not used in the above rules, so keep a simple TestCase here
        try:
            urlrewriter.rewrite("logilab.fr", "/whatever", req)
            self.fail("redirect exception expected")
        except Redirect as ex:
            self.assertEqual(ex.location, "http://www.logilab.fr/whatever")
        self.assertEqual(
            urlrewriter.rewrite("www.logilab.fr", "/whatever", req), "/whatever"
        )
        self.assertEqual(
            urlrewriter.rewrite("www.logilab.fr", "/json/bla", req), "/json/bla"
        )
        self.assertEqual(
            urlrewriter.rewrite("abcd.logilab.fr", "/json/bla", req), "/json/bla"
        )
        self.assertEqual(
            urlrewriter.rewrite("abcd.logilab.fr", "/data/bla", req), "/data/bla"
        )
        self.assertEqual(
            urlrewriter.rewrite("abcd.logilab.fr", "/whatever", req), "/m_abcd/whatever"
        )
        self.assertEqual(urlrewriter.rewrite("abcd.fr", "/whatever", req), "/whatever")


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Unit tests for url publishing service"""

import re

from logilab.common.testlib import unittest_main

from cubicweb.rset import ResultSet
from cubicweb_web import NotFound, Redirect, views
from cubicweb_web.devtools.testlib import WebCWTC, FakeRequest
from cubicweb_web.views.urlrewrite import SimpleReqRewriter


class URLPublisherCWTC(WebCWTC):
    """test suite for QSPreProcessor"""

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            self.create_user(cnx, "ÿsaÿe")
            b = cnx.create_entity("BlogEntry", title="hell'o", content="blabla")
            # take care: Tag's name normalized to lower case
            c = cnx.create_entity("Tag", name="yo")
            cnx.execute(
                "SET C tags B WHERE C eid %(c)s, B eid %(b)s", {"c": c.eid, "b": b.eid}
            )
            cnx.commit()

    def process(self, req, url):
        return self.app.url_resolver.process(req, url)

    def test_raw_path(self):
        """tests raw path resolution'"""
        with self.admin_access.web_request() as req:
            self.assertEqual(self.process(req, "view"), ("view", None))
            self.assertEqual(self.process(req, "edit"), ("edit", None))
            self.assertRaises(NotFound, self.process, req, "whatever")

    def test_eid_path(self):
        """tests eid path resolution"""
        with self.admin_access.web_request() as req:
            self.assertIsInstance(self.process(req, "123")[1], ResultSet)
            self.assertEqual(len(self.process(req, "123")[1]), 1)
            self.assertRaises(NotFound, self.process, req, "123/345")
            self.assertRaises(NotFound, self.process, req, "not_eid")

    def test_rest_path_etype(self):
        """tests the rest path resolution"""
        with self.admin_access.web_request() as req:
            ctrl, rset = self.process(req, "CWEType")
            self.assertEqual(ctrl, "view")
            self.assertEqual(rset.description[0][0], "CWEType")
            self.assertEqual(
                "Any X,AA,AB ORDERBY AB WHERE X is_instance_of CWEType, "
                "X modification_date AA, X name AB",
                rset.printable_rql(),
            )
            self.assertEqual(req.form["vid"], "sameetypelist")

    def test_rest_path_by_attr(self):
        with self.admin_access.web_request() as req:
            ctrl, rset = self.process(req, "CWUser/login/admin")
            self.assertEqual(ctrl, "view")
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset.description[0][0], "CWUser")
            self.assertEqual(
                "Any X,AA,AB,AC,AD WHERE X is_instance_of CWUser, "
                "X firstname AA, X login AB, X modification_date AC, "
                'X surname AD, X login "admin"',
                rset.printable_rql(),
            )

    def test_rest_path_unique_attr(self):
        with self.admin_access.web_request() as req:
            ctrl, rset = self.process(req, "cwuser/admin")
            self.assertEqual(ctrl, "view")
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset.description[0][0], "CWUser")
            self.assertEqual(
                "Any X,AA,AB,AC,AD WHERE X is_instance_of CWUser, "
                "X firstname AA, X login AB, X modification_date AC, "
                'X surname AD, X login "admin"',
                rset.printable_rql(),
            )
            self.assertEqual(req.form["vid"], "primary")

    def test_rest_path_eid(self):
        with self.admin_access.web_request() as req:
            ctrl, rset = self.process(req, "cwuser/eid/%s" % req.user.eid)
            self.assertEqual(ctrl, "view")
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset.description[0][0], "CWUser")
            self.assertEqual(
                "Any X,AA,AB,AC,AD WHERE X is_instance_of CWUser, "
                "X firstname AA, X login AB, X modification_date AC, "
                "X surname AD, X eid %s" % rset[0][0],
                rset.printable_rql(),
            )

    def test_rest_path_non_ascii_paths(self):
        with self.admin_access.web_request() as req:
            ctrl, rset = self.process(req, "CWUser/login/%C3%BFsa%C3%BFe")
            self.assertEqual(ctrl, "view")
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset.description[0][0], "CWUser")
            self.assertEqual(
                "Any X,AA,AB,AC,AD WHERE X is_instance_of CWUser, "
                "X firstname AA, X login AB, X modification_date AC, "
                'X surname AD, X login "\xffsa\xffe"',
                rset.printable_rql(),
            )

    def test_rest_path_quoted_paths(self):
        with self.admin_access.web_request() as req:
            ctrl, rset = self.process(req, "BlogEntry/title/hell%27o")
            self.assertEqual(ctrl, "view")
            self.assertEqual(len(rset), 1)
            self.assertEqual(rset.description[0][0], "BlogEntry")
            self.assertEqual(
                "Any X,AA,AB,AC WHERE X is_instance_of BlogEntry, "
                "X creation_date AA, X modification_date AB, X title AC, "
                'X title "hell\'o"',
                rset.printable_rql(),
            )

    def test_rest_path_use_vid_from_rset(self):
        with self.admin_access.web_request(
            headers={"Accept": "application/rdf+xml"}
        ) as req:
            views.VID_BY_MIMETYPE["application/rdf+xml"] = "rdf"
            try:
                ctrl, rset = self.process(req, "CWEType")
            finally:
                views.VID_BY_MIMETYPE.pop("application/rdf+xml")
            self.assertEqual(req.form["vid"], "rdf")

    def test_rest_path_errors(self):
        with self.admin_access.web_request() as req:
            self.assertRaises(NotFound, self.process, req, "CWUser/eid/30000")
            self.assertRaises(NotFound, self.process, req, "Workcases")
            self.assertRaises(
                NotFound, self.process, req, "CWUser/inexistant_attribute/joe"
            )

    def test_action_path(self):
        """tests the action path resolution"""
        with self.admin_access.web_request() as req:
            self.assertRaises(Redirect, self.process, req, "1/edit")
            self.assertRaises(Redirect, self.process, req, "Tag/name/yo/edit")
            self.assertRaises(Redirect, self.process, req, "Tag/yo/edit")
            self.assertRaises(NotFound, self.process, req, "view/edit")
            self.assertRaises(NotFound, self.process, req, "1/non_action")
            self.assertRaises(
                NotFound, self.process, req, "CWUser/login/admin/non_action"
            )

    def test_regexp_path(self):
        """tests the regexp path resolution"""
        with self.admin_access.web_request() as req:
            ctrl, rset = self.process(req, "add/Task")
            self.assertEqual(ctrl, "view")
            self.assertEqual(rset, None)
            self.assertEqual(req.form, {"etype": "Task", "vid": "creation"})
            self.assertRaises(NotFound, self.process, req, "add/foo/bar")

    def test_nonascii_path(self):
        oldrules = SimpleReqRewriter.rules
        SimpleReqRewriter.rules = [(re.compile(r"/\w+", re.U), dict(vid="foo"))]
        with self.admin_access.web_request() as req:
            try:
                path = str(FakeRequest().url_quote("été"))
                ctrl, rset = self.process(req, path)
                self.assertEqual(rset, None)
                self.assertEqual(req.form, {"vid": "foo"})
            finally:
                SimpleReqRewriter.rules = oldrules


if __name__ == "__main__":
    unittest_main()



from urllib.parse import urlparse, urljoin

from cubicweb.devtools import BASE_URL
from cubicweb_web.devtools.testlib import PyramidWebCWTC


def simple_test_view(request):
    return request.response


class AnonymousDisallowedPredicatesTC(PyramidWebCWTC):
    anonymous_allowed = False

    def includeme(self, config):
        config.add_route("view-entity", r"/{eid}", has_cw_permission="read")
        config.add_view(simple_test_view, route_name="view-entity")
        config.registry.settings["pyramid.csrf_trusted_origins"] = ["example.com"]

    def setUp(self):
        super(AnonymousDisallowedPredicatesTC, self).setUp()
        with self.admin_access.repo_cnx() as cnx:
            self.eid = self.create_user(cnx, "Fernestin").eid
            cnx.commit()

    def test_view_entity_without_connection(self):
        self.webapp.get(urljoin(BASE_URL, str(self.eid)), status=303)

    def test_view_entity_with_connection(self):
        self.login(user="Fernestin", password="Fernestin")
        self.webapp.get(urljoin(BASE_URL, str(self.eid)), status=200)


class PredicatesTC(PyramidWebCWTC):
    def includeme(self, config):
        config.route_prefix = urlparse(BASE_URL).path
        # to test route predicate
        config.add_route(
            "permission-update-route",
            r"/{eid}/test-permission-update-route",
            has_cw_permission="update",
        )
        config.add_view(simple_test_view, route_name="permission-update-route")
        # to test view predicate
        config.add_route(
            "permission-update-view", r"/{eid}/test-permission-update-view"
        )
        config.add_view(
            simple_test_view,
            route_name="permission-update-view",
            has_cw_permission="update",
        )
        config.registry.settings["pyramid.csrf_trusted_origins"] = ["example.com"]

    def test_has_cw_permission_route_predicate(self):
        with self.admin_access.repo_cnx() as cnx:
            eid = self.create_user(cnx, "Fernestin").eid
            cnx.commit()
        res = self.webapp.get(
            urljoin(BASE_URL, f"{eid}/test-permission-update-route"), status="*"
        )
        self.assertEqual(res.status_int, 404)
        self.webapp.post(
            urljoin(BASE_URL, "login"),
            {"__login": self.admlogin, "__password": self.admpassword},
        )
        res = self.webapp.get(
            urljoin(BASE_URL, f"{eid}/test-permission-update-route"), status="*"
        )
        self.assertEqual(res.status_int, 200)

    def test_has_cw_permission_view_predicate(self):
        with self.admin_access.repo_cnx() as cnx:
            eid = self.create_user(cnx, "Fernestin").eid
            cnx.commit()
        res = self.webapp.get(
            urljoin(BASE_URL, f"{eid}/test-permission-update-view"), status="*"
        )
        self.assertEqual(res.status_int, 404)
        self.webapp.login(user=self.admlogin, password=self.admpassword)
        res = self.webapp.get(
            urljoin(BASE_URL, f"{eid}/test-permission-update-view"), status="*"
        )
        self.assertEqual(res.status_int, 200)



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""some basic stuff to build dot generated graph images"""

import codecs
import os
import tempfile

from logilab.common.graph import GraphGenerator, DotBackend

from cubicweb_web.view import EntityView


class DotGraphView(EntityView):
    __abstract__ = True
    backend_class = DotBackend
    backend_kwargs = {"ratio": "compress", "size": "30,10"}

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        visitor = self.build_visitor(entity)
        prophdlr = self.build_dotpropshandler()
        graphname = "dotgraph%s" % str(entity.eid)
        generator = GraphGenerator(
            self.backend_class(graphname, None, **self.backend_kwargs)
        )
        # image file
        fd, tmpfile = tempfile.mkstemp(".svg")
        os.close(fd)
        generator.generate(visitor, prophdlr, tmpfile)
        with codecs.open(tmpfile, "rb", encoding="utf-8") as svgfile:
            self.w(svgfile.read())

    def build_visitor(self, entity):
        raise NotImplementedError

    def build_dotpropshandler(self):
        return DotPropsHandler(self._cw)


class DotPropsHandler:
    def __init__(self, req):
        self._ = req._

    def node_properties(self, entity):
        """return default DOT drawing options for a state or transition"""
        return {
            "label": entity.dc_long_title(),
            "href": entity.absolute_url(),
            "fontname": "Courier",
            "fontsize": 10,
            "shape": "box",
        }

    def edge_properties(self, transition, fromstate, tostate):
        return {"label": "", "dir": "forward", "color": "black", "style": "filled"}



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb.devtools.htmlparser import XMLValidator

from cubicweb_web.devtools.testlib import WebCWTC


class LogFormTemplateCWTC(WebCWTC):
    def _login_labels(self):
        valid = self.content_type_validators.get("text/html", XMLValidator)()
        req = self.requestcls(self.vreg, url="login")
        page = valid.parse_string(self.vreg["views"].main_template(req, "login"))
        return page.find_tag("label")

    def test_label(self):
        self.set_option("allow-email-login", "yes")
        self.assertEqual(self._login_labels(), ["login or email", "password"])
        self.set_option("allow-email-login", "no")
        self.assertEqual(self._login_labels(), ["login", "password"])

    def test_display_message(self):
        with self.admin_access.web_request() as req:
            req.set_message("houla hop")
            page = self.view("logform", req=req, id="loginBox", klass="", template=None)
            self.assertIn("houla hop", page.raw_text)


class MainNoTopTemplateCWTC(WebCWTC):
    def test_valid_xhtml(self):
        with self.admin_access.web_request() as req:
            self.view("index", template="main-no-top", req=req)


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2023-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from cubicweb.predicates import (
    match_form_params,
)

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.predicates import paginated_rset
from logilab.common.testlib import unittest_main


class MatchFormParamsCWTC(WebCWTC):
    """tests for match_form_params predicate"""

    def test_keyonly_match(self):
        """test standard usage: ``match_form_params('param1', 'param2')``

        ``param1`` and ``param2`` must be specified in request's form.
        """
        web_request = self.admin_access.web_request
        vid_selector = match_form_params("vid")
        vid_subvid_selector = match_form_params("vid", "subvid")
        # no parameter => KO,KO
        with web_request() as req:
            self.assertEqual(vid_selector(None, req), 0)
            self.assertEqual(vid_subvid_selector(None, req), 0)
        # one expected parameter found => OK,KO
        with web_request(vid="foo") as req:
            self.assertEqual(vid_selector(None, req), 1)
            self.assertEqual(vid_subvid_selector(None, req), 0)
        # all expected parameters found => OK,OK
        with web_request(vid="foo", subvid="bar") as req:
            self.assertEqual(vid_selector(None, req), 1)
            self.assertEqual(vid_subvid_selector(None, req), 2)

    def test_keyvalue_match_one_parameter(self):
        """test dict usage: ``match_form_params(param1=value1)``

        ``param1`` must be specified in the request's form and its value
        must be ``value1``.
        """
        web_request = self.admin_access.web_request
        # test both positional and named parameters
        vid_selector = match_form_params(vid="foo")
        # no parameter => should fail
        with web_request() as req:
            self.assertEqual(vid_selector(None, req), 0)
        # expected parameter found with expected value => OK
        with web_request(vid="foo", subvid="bar") as req:
            self.assertEqual(vid_selector(None, req), 1)
        # expected parameter found but value is incorrect => KO
        with web_request(vid="bar") as req:
            self.assertEqual(vid_selector(None, req), 0)

    def test_keyvalue_match_two_parameters(self):
        """test dict usage: ``match_form_params(param1=value1, param2=value2)``

        ``param1`` and ``param2`` must be specified in the request's form and
        their respective value must be ``value1`` and ``value2``.
        """
        web_request = self.admin_access.web_request
        vid_subvid_selector = match_form_params(vid="list", subvid="tsearch")
        # missing one expected parameter => KO
        with web_request(vid="list") as req:
            self.assertEqual(vid_subvid_selector(None, req), 0)
        # expected parameters found but values are incorrect => KO
        with web_request(vid="list", subvid="foo") as req:
            self.assertEqual(vid_subvid_selector(None, req), 0)
        # expected parameters found and values are correct => OK
        with web_request(vid="list", subvid="tsearch") as req:
            self.assertEqual(vid_subvid_selector(None, req), 2)

    def test_keyvalue_multiple_match(self):
        """test dict usage with multiple values

        i.e. as in ``match_form_params(param1=('value1', 'value2'))``

        ``param1`` must be specified in the request's form and its value
        must be either ``value1`` or ``value2``.
        """
        web_request = self.admin_access.web_request
        vid_subvid_selector = match_form_params(
            vid="list", subvid=("tsearch", "listitem")
        )
        # expected parameters found and values correct => OK
        with web_request(vid="list", subvid="tsearch") as req:
            self.assertEqual(vid_subvid_selector(None, req), 2)
        with web_request(vid="list", subvid="listitem") as req:
            self.assertEqual(vid_subvid_selector(None, req), 2)
        # expected parameters found but values are incorrect => OK
        with web_request(vid="list", subvid="foo") as req:
            self.assertEqual(vid_subvid_selector(None, req), 0)

    def test_invalid_calls(self):
        """checks invalid calls raise a ValueError"""
        # mixing named and positional arguments should fail
        with self.assertRaises(ValueError) as cm:
            match_form_params("list", x="1", y="2")
        self.assertEqual(
            str(cm.exception),
            "match_form_params() can't be called with both "
            "positional and named arguments",
        )
        # using a dict as first and unique argument should fail
        with self.assertRaises(ValueError) as cm:
            match_form_params({"x": 1})
        self.assertEqual(
            str(cm.exception),
            "match_form_params() positional arguments must be strings",
        )


class PaginatedCWTC(WebCWTC):
    """tests for paginated_rset predicate"""

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            for i in range(30):
                cnx.create_entity("CWGroup", name="group%d" % i)
            cnx.commit()

    def test_paginated_rset(self):
        default_nb_pages = 1
        web_request = self.admin_access.web_request
        with web_request() as req:
            rset = req.execute("Any G WHERE G is CWGroup")
        self.assertEqual(len(rset), 34)
        with web_request(vid="list", page_size="10") as req:
            self.assertEqual(paginated_rset()(None, req, rset), default_nb_pages)
        with web_request(vid="list", page_size="20") as req:
            self.assertEqual(paginated_rset()(None, req, rset), default_nb_pages)
        with web_request(vid="list", page_size="50") as req:
            self.assertEqual(paginated_rset()(None, req, rset), 0)
        with web_request(vid="list", page_size="10/") as req:
            self.assertEqual(paginated_rset()(None, req, rset), 0)
        with web_request(vid="list", page_size=".1") as req:
            self.assertEqual(paginated_rset()(None, req, rset), 0)
        with web_request(vid="list", page_size="not_an_int") as req:
            self.assertEqual(paginated_rset()(None, req, rset), 0)


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""basic plot views"""

from logilab.common.registry import objectify_predicate

from cubicweb.predicates import multi_columns_rset
from cubicweb.utils import UStringIO
from cubicweb_web.views import baseviews


@objectify_predicate
def all_columns_are_numbers(cls, req, rset=None, *args, **kwargs):
    """accept result set with at least one line and two columns of result
    all columns after second must be of numerical types"""
    for etype in rset.description[0]:
        if etype not in ("Int", "BigInt", "Float"):
            return 0
    return 1


@objectify_predicate
def second_column_is_number(cls, req, rset=None, *args, **kwargs):
    etype = rset.description[0][1]
    if etype not in ("Int", "BigInt", "Float"):
        return 0
    return 1


@objectify_predicate
def columns_are_date_then_numbers(cls, req, rset=None, *args, **kwargs):
    etypes = rset.description[0]
    if etypes[0] not in ("Date", "Datetime", "TZDatetime"):
        return 0
    for etype in etypes[1:]:
        if etype not in ("Int", "BigInt", "Float"):
            return 0
    return 1


def filterout_nulls(abscissa, plot):
    filtered = []
    for x, y in zip(abscissa, plot):
        if x is None or y is None:
            continue
        filtered.append((x, y))
    return sorted(filtered)


class PlotWidget:
    # XXX refactor with cubicweb_web.views.htmlwidgets.HtmlWidget
    def _initialize_stream(self, w=None):
        if w:
            self.w = w
        else:
            self._stream = UStringIO()
            self.w = self._stream.write

    def render(self, *args, **kwargs):
        w = kwargs.pop("w", None)
        self._initialize_stream(w)
        self._render(*args, **kwargs)
        if w is None:
            return self._stream.getvalue()

    def _render(self, *args, **kwargs):
        raise NotImplementedError


try:
    from GChartWrapper import Pie, Pie3D
except ImportError:
    pass
else:

    class PieChartWidget(PlotWidget):
        def __init__(self, labels, values, pieclass=Pie, title=None):
            self.labels = labels
            self.values = values
            self.pieclass = pieclass
            self.title = title

        def _render(self, width=None, height=None):
            piechart = self.pieclass(self.values)
            piechart.label(*self.labels)
            if width is not None:
                height = height or width
                piechart.size(width, height)
            if self.title:
                piechart.title(self.title)
            self.w('<img src="%s" />', piechart.url)

    class PieChartView(baseviews.AnyRsetView):
        __regid__ = "piechart"
        pieclass = Pie
        paginable = False

        __select__ = multi_columns_rset() & second_column_is_number()

        def _guess_vid(self, row):
            etype = self.cw_rset.description[row][0]
            if self._cw.vreg.schema.eschema(etype).final:
                return "final"
            return "textincontext"

        def call(self, title=None, width=None, height=None):
            labels = []
            values = []
            for rowidx, (_, value) in enumerate(self.cw_rset):
                if value is not None:
                    vid = self._guess_vid(rowidx)
                    label = "{}: {}".format(
                        self._cw.view(vid, self.cw_rset, row=rowidx, col=0),
                        value,
                    )
                    labels.append(label.encode(self._cw.encoding))
                    values.append(value)
            pie = PieChartWidget(labels, values, pieclass=self.pieclass, title=title)
            if width is not None:
                height = height or width
            pie.render(width, height, w=self.w)

    class PieChart3DView(PieChartView):
        __regid__ = "piechart3D"
        pieclass = Pie3D



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""provide class to do Apache rewrite rules'job inside cubicweb (though functionnalities
are much more limited for the moment)

"""


from re import compile

from cubicweb_web import Redirect
from cubicweb_web.component import Component


class RewriteCond:
    def __init__(self, condition, match="host", rules=(), action="rewrite"):
        self.condition = compile(condition)
        assert match in ("host", "path"), match
        self.match_part = match
        self.rules = []
        for rule, replace in rules:
            rulergx = compile(rule)
            self.rules.append((rulergx, replace))
        assert action in ("rewrite", "redirect", "stop"), action
        self.process = getattr(self, "action_%s" % action)

    def match(self, **kwargs):
        self._match = self.condition.match(kwargs[self.match_part])
        return self._match is not None

    def action_rewrite(self, path):
        for rgx, replace in self.rules:
            if not rgx.match(path) is None:
                matchdict = self._match.groupdict() or None
                if matchdict is not None:
                    replace = replace % matchdict
                return rgx.sub(replace, path)
        return path

    def action_redirect(self, path):
        url = self.action_rewrite(path)
        raise Redirect(url)

    def action_stop(self, path):
        return path


class ApacheURLRewrite(Component):
    """inherit from this class with actual rules to activate apache style rewriting

    rules should have the form :

    [('condition pattern 1', [('rule1 pattern', 'replace expression'),
                              ('rule2 pattern', 'replace expression')],
     ('condition pattern 2', [('rule1 pattern', 'replace expression'),
                              ('rule2 pattern', 'replace expression')]
    ]

    for instance the equivalent of the following apache rules:

        RewriteCond %{HTTP_HOST} ^logilab\\.fr
        RewriteRule ^/(.*) http://www.logilab.fr/$1 [L,R=301]

        RewriteCond %{HTTP_HOST} ^www\\.logilab\\.fr
        RewriteRule ^/(.*) http://localhost:8080/$1 [L,P]

        RewriteCond %{HTTP_HOST} ^(.+)\\.logilab\\.fr
        RewriteRule ^/(data/.*) http://localhost:8080/$1 [L,P]
        RewriteRule ^/(json.*) http://localhost:8080/$1 [L,P]
        RewriteRule ^/(.*) http://localhost:8080/m_%1/$1 [L,P]

    could be written (considering that no "host rewritting" is necessary):

      class MyAppRules(ApacheURLRewrite):
        rules = [
          RewriteCond('logilab\\.fr', match='host',
                      rules=[('/(.*)', r'http://www.logilab.fr/\1')],
                      action='redirect'),
          RewriteCond('(www)\\.logilab\\.fr', match='host', action='stop'),
          RewriteCond('/(data|json)/', match='path', action='stop'),
          RewriteCond('(?P<cat>.*)\\.logilab\\.fr', match='host',
                      rules=[('/(.*)', r'/m_%(cat)s/\1')]),
        ]
    """

    __abstract__ = True
    __regid__ = "urlrewriter"
    rules = []

    def get_rules(self, req):
        return self.rules

    def rewrite(self, host, path, req):
        for cond in self.get_rules(req):
            if cond.match(host=host, path=path):
                return cond.process(path)
        return path



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Unit tests for cw.web.views.magicsearch"""

from contextlib import contextmanager

from logilab.common.testlib import unittest_main
from rql import BadRQLQuery, RQLSyntaxError

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views.magicsearch import (
    QSPreProcessor,
    QueryTranslator,
)

translations = {
    "CWUser": "Utilisateur",
    "EmailAddress": "Adresse",
    "name": "nom",
    "alias": "nom",
    "surname": "nom",
    "firstname": "prÃ©nom",
    "state": "Ã©tat",
    "address": "adresse",
    "use_email": "adel",
}


def _translate(msgid):
    return translations.get(msgid, msgid)


def _ctxtranslate(ctx, msgid):
    return _translate(msgid)


class QueryTranslatorCWTC(WebCWTC):
    """test suite for QueryTranslatorTC"""

    @contextmanager
    def proc(self):
        with self.admin_access.web_request() as req:
            self.vreg.config.translations = {"en": (_translate, _ctxtranslate)}
            proc = self.vreg["components"].select("magicsearch", req)
            proc = [p for p in proc.processors if isinstance(p, QueryTranslator)][0]
            yield proc

    def test_basic_translations(self):
        """tests basic translations (no ambiguities)"""
        with self.proc() as proc:
            rql = "Any C WHERE C is Adresse, P adel C, C adresse 'Logilab'"
            (rql,) = proc.preprocess_query(rql)
            self.assertEqual(
                rql, 'Any C WHERE C is EmailAddress, P use_email C, C address "Logilab"'
            )

    def test_ambiguous_translations(self):
        """tests possibly ambiguous translations"""
        with self.proc() as proc:
            rql = "Any P WHERE P adel C, C is EmailAddress, C nom 'Logilab'"
            (rql,) = proc.preprocess_query(rql)
            self.assertEqual(
                rql, 'Any P WHERE P use_email C, C is EmailAddress, C alias "Logilab"'
            )
            rql = "Any P WHERE P is Utilisateur, P adel C, P nom 'Smith'"
            (rql,) = proc.preprocess_query(rql)
            self.assertEqual(
                rql, 'Any P WHERE P is CWUser, P use_email C, P surname "Smith"'
            )


class QSPreProcessorCWTC(WebCWTC):
    """test suite for QSPreProcessor"""

    @contextmanager
    def proc(self):
        self.vreg.config.translations = {"en": (_translate, _ctxtranslate)}
        with self.admin_access.web_request() as req:
            proc = self.vreg["components"].select("magicsearch", req)
            proc = [p for p in proc.processors if isinstance(p, QSPreProcessor)][0]
            proc._cw = req
            yield proc

    def test_entity_translation(self):
        """tests QSPreProcessor._get_entity_name()"""
        with self.proc() as proc:
            translate = proc._get_entity_type
            self.assertEqual(translate("EmailAddress"), "EmailAddress")
            self.assertEqual(translate("emailaddress"), "EmailAddress")
            self.assertEqual(translate("Adresse"), "EmailAddress")
            self.assertEqual(translate("adresse"), "EmailAddress")
            self.assertRaises(BadRQLQuery, translate, "whatever")

    def test_attribute_translation(self):
        """tests QSPreProcessor._get_attribute_name"""
        with self.proc() as proc:
            translate = proc._get_attribute_name
            eschema = self.schema.entity_schema_for("CWUser")
            self.assertEqual(translate("prÃ©nom", eschema), "firstname")
            self.assertEqual(translate("nom", eschema), "surname")
            eschema = self.schema.entity_schema_for("EmailAddress")
            self.assertEqual(translate("adresse", eschema), "address")
            self.assertEqual(translate("nom", eschema), "alias")
            # should fail if the name is not an attribute for the given entity schema
            self.assertRaises(BadRQLQuery, translate, "whatever", eschema)
            self.assertRaises(BadRQLQuery, translate, "prÃ©nom", eschema)

    def test_one_word_query(self):
        """tests the 'one word shortcut queries'"""
        with self.proc() as proc:
            transform = proc._one_word_query
            self.assertEqual(
                transform("123"), ("Any X WHERE X eid %(x)s", {"x": 123}, "x")
            )
            self.assertEqual(transform("CWUser"), ("CWUser C",))
            self.assertEqual(transform("Utilisateur"), ("CWUser C",))
            self.assertEqual(transform("Adresse"), ("EmailAddress E",))
            self.assertEqual(transform("adresse"), ("EmailAddress E",))
            self.assertRaises(BadRQLQuery, transform, "Workcases")

    def test_two_words_query(self):
        """tests the 'two words shortcut queries'"""
        with self.proc() as proc:
            transform = proc._two_words_query
            self.assertEqual(transform("CWUser", "E"), ("CWUser E",))
            self.assertEqual(
                transform("CWUser", "Smith"),
                (
                    "CWUser C ORDERBY FTIRANK(C) DESC WHERE C has_text %(text)s",
                    {"text": "Smith"},
                ),
            )
            self.assertEqual(
                transform("utilisateur", "Smith"),
                (
                    "CWUser C ORDERBY FTIRANK(C) DESC WHERE C has_text %(text)s",
                    {"text": "Smith"},
                ),
            )
            self.assertEqual(
                transform("adresse", "Logilab"),
                (
                    "EmailAddress E ORDERBY FTIRANK(E) DESC WHERE E has_text %(text)s",
                    {"text": "Logilab"},
                ),
            )
            self.assertEqual(
                transform("adresse", "Logi%"),
                ("EmailAddress E WHERE E alias LIKE %(text)s", {"text": "Logi%"}),
            )
            self.assertRaises(BadRQLQuery, transform, "pers", "taratata")

    def test_three_words_query(self):
        """tests the 'three words shortcut queries'"""
        with self.proc() as proc:
            transform = proc._three_words_query
            self.assertEqual(
                transform("utilisateur", "prÃ©nom", "cubicweb"),
                ("CWUser C WHERE C firstname %(text)s", {"text": "cubicweb"}),
            )
            self.assertEqual(
                transform("utilisateur", "nom", "cubicweb"),
                ("CWUser C WHERE C surname %(text)s", {"text": "cubicweb"}),
            )
            self.assertEqual(
                transform("adresse", "nom", "cubicweb"),
                ("EmailAddress E WHERE E alias %(text)s", {"text": "cubicweb"}),
            )
            self.assertEqual(
                transform("EmailAddress", "nom", "cubicweb"),
                ("EmailAddress E WHERE E alias %(text)s", {"text": "cubicweb"}),
            )
            self.assertEqual(
                transform("utilisateur", "prÃ©nom", "cubicweb%"),
                ("CWUser C WHERE C firstname LIKE %(text)s", {"text": "cubicweb%"}),
            )
            # expanded shortcuts
            self.assertEqual(
                transform("CWUser", "use_email", "Logilab"),
                (
                    "CWUser C ORDERBY FTIRANK(C1) DESC WHERE C use_email C1, C1 has_text %(text)s",
                    {"text": "Logilab"},
                ),
            )
            self.assertEqual(
                transform("CWUser", "use_email", "%Logilab"),
                (
                    "CWUser C WHERE C use_email C1, C1 alias LIKE %(text)s",
                    {"text": "%Logilab"},
                ),
            )
            self.assertRaises(BadRQLQuery, transform, "word1", "word2", "word3")

    def test_quoted_queries(self):
        """tests how quoted queries are handled"""
        queries = [
            (
                'Adresse "My own EmailAddress"',
                (
                    "EmailAddress E ORDERBY FTIRANK(E) DESC WHERE E has_text %(text)s",
                    {"text": "My own EmailAddress"},
                ),
            ),
            (
                'Utilisateur prÃ©nom "Jean Paul"',
                ("CWUser C WHERE C firstname %(text)s", {"text": "Jean Paul"}),
            ),
            (
                'Utilisateur firstname "Jean Paul"',
                ("CWUser C WHERE C firstname %(text)s", {"text": "Jean Paul"}),
            ),
            (
                'CWUser firstname "Jean Paul"',
                ("CWUser C WHERE C firstname %(text)s", {"text": "Jean Paul"}),
            ),
        ]
        with self.proc() as proc:
            transform = proc._quoted_words_query
            for query, expected in queries:
                self.assertEqual(transform(query), expected)
            self.assertRaises(BadRQLQuery, transform, "unquoted rql")
            self.assertRaises(BadRQLQuery, transform, 'pers "Jean Paul"')
            self.assertRaises(
                BadRQLQuery, transform, 'CWUser firstname other "Jean Paul"'
            )

    def test_process_query(self):
        """tests how queries are processed"""
        queries = [
            ("Utilisateur", ("CWUser C",)),
            ("Utilisateur P", ("CWUser P",)),
            (
                "Utilisateur cubicweb",
                (
                    "CWUser C ORDERBY FTIRANK(C) DESC WHERE C has_text %(text)s",
                    {"text": "cubicweb"},
                ),
            ),
            (
                "CWUser prÃ©nom cubicweb",
                (
                    "CWUser C WHERE C firstname %(text)s",
                    {"text": "cubicweb"},
                ),
            ),
        ]
        with self.proc() as proc:
            for query, expected in queries:
                self.assertEqual(proc.preprocess_query(query), expected)
            self.assertRaises(
                BadRQLQuery, proc.preprocess_query, "Any X WHERE X is Something"
            )


# Processor Chains tests ############################################


class ProcessorChainCWTC(WebCWTC):
    """test suite for magic_search's processor chains"""

    @contextmanager
    def proc(self):
        self.vreg.config.translations = {"en": (_translate, _ctxtranslate)}
        with self.admin_access.web_request() as req:
            proc = self.vreg["components"].select("magicsearch", req)
            yield proc

    def test_main_preprocessor_chain(self):
        """tests QUERY_PROCESSOR"""
        queries = [
            (
                "foo",
                (
                    "Any X ORDERBY FTIRANK(X) DESC WHERE X has_text %(text)s",
                    {"text": "foo"},
                ),
            ),
            # XXX this sounds like a language translator test...
            # and it fails
            (
                "Utilisateur Smith",
                (
                    "CWUser C ORDERBY FTIRANK(C) DESC WHERE C has_text %(text)s",
                    {"text": "Smith"},
                ),
            ),
            (
                "utilisateur nom Smith",
                ("CWUser C WHERE C surname %(text)s", {"text": "Smith"}),
            ),
            (
                'Any P WHERE P is Utilisateur, P nom "Smith"',
                ('Any P WHERE P is CWUser, P surname "Smith"', None),
            ),
        ]
        with self.proc() as proc:
            for query, expected in queries:
                rset = proc.process_query(query)
                self.assertEqual((rset.rql, rset.args), expected)

    def test_accentuated_fulltext(self):
        """we must be able to type accentuated characters in the search field"""
        with self.proc() as proc:
            rset = proc.process_query("écrire")
            self.assertEqual(
                rset.rql, "Any X ORDERBY FTIRANK(X) DESC WHERE X has_text %(text)s"
            )
            self.assertEqual(rset.args, {"text": "écrire"})

    def test_explicit_component(self):
        with self.proc() as proc:
            self.assertRaises(
                RQLSyntaxError,
                proc.process_query,
                'rql: CWUser E WHERE E noattr "Smith",',
            )
            self.assertRaises(
                BadRQLQuery, proc.process_query, 'rql: CWUser E WHERE E noattr "Smith"'
            )
            rset = proc.process_query("text: utilisateur Smith")
            self.assertEqual(
                rset.rql, "Any X ORDERBY FTIRANK(X) DESC WHERE X has_text %(text)s"
            )
            self.assertEqual(rset.args, {"text": "utilisateur Smith"})


if __name__ == "__main__":
    unittest_main()



import errno
import os
import tempfile
from os.path import join, dirname
from shutil import rmtree
from unittest import TestCase, main

from cubicweb_web.propertysheet import PropertySheet, lazystr

DATADIR = join(dirname(__file__), "data")


class PropertySheetTC(TestCase):
    def setUp(self):
        uicache = join(DATADIR, "uicache")
        try:
            os.makedirs(uicache)
        except OSError as err:
            if err.errno != errno.EEXIST:
                raise
        self.cachedir = tempfile.mkdtemp(dir=uicache)

    def tearDown(self):
        rmtree(self.cachedir)

    def data(self, filename):
        return join(DATADIR, filename)

    def test(self):
        ps = PropertySheet(self.cachedir, datadir_url="http://cwtest.com")
        ps.load(self.data("sheet1.py"))
        ps.load(self.data("sheet2.py"))
        # defined by sheet1
        self.assertEqual(ps["logo"], "http://cwtest.com/logo.png")
        # defined by sheet1, overriden by sheet2
        self.assertEqual(ps["bgcolor"], "#FFFFFF")
        # defined by sheet2
        self.assertEqual(ps["fontcolor"], "black")
        # defined by sheet1, extended by sheet2
        self.assertEqual(
            ps["stylesheets"],
            ["http://cwtest.com/cubicweb.css", "http://cwtest.com/mycube.css"],
        )
        # lazy string defined by sheet1
        self.assertIsInstance(ps["lazy"], lazystr)
        self.assertEqual(str(ps["lazy"]), "#FFFFFF")
        # test compilation
        self.assertEqual(
            ps.compile("a {bgcolor: %(bgcolor)s; size: 1%;}"),
            "a {bgcolor: #FFFFFF; size: 1%;}",
        )
        self.assertEqual(ps.process_resource(DATADIR, "pouet.css"), self.cachedir)
        self.assertFalse(ps.need_reload())
        os.utime(self.data("sheet1.py"), None)
        self.assertTrue(ps.need_reload())
        ps.reload()
        self.assertFalse(ps.need_reload())
        ps.process_resource(DATADIR, "pouet.css")  # put in cache
        os.utime(self.data("pouet.css"), None)
        self.assertFalse(ps.need_reload())

    def test_chmod(self):
        ps = PropertySheet(self.cachedir, datadir_url="http://cwtest.com")
        ps.load(self.data("sheet1.py"))
        rdir = ps.process_resource(DATADIR, "pouet.css")
        expected_mode = os.stat(self.data("sheet1.py")).st_mode
        got_mode = os.stat(join(rdir, "pouet.css")).st_mode
        self.assertEqual(got_mode, expected_mode)


if __name__ == "__main__":
    main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
import glob
import os
import os.path as osp
from contextlib import contextmanager

from logilab.common import tempattr
from logilab.common.testlib import Tags

from cubicweb_web.utils import HTMLHead

from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views.staticcontrollers import ConcatFilesHandler


class staticfilespublishermixin:
    @contextmanager
    def _publish_static_files(self, url, header={}):
        with self.admin_access.web_request(headers=header) as req:
            req._url = url
            self.app_handle_request(req)
            yield req


class StaticControllerCacheCWTC(staticfilespublishermixin, WebCWTC):
    tags = WebCWTC.tags | Tags("static_controller", "cache", "http")

    def test_static_file_are_cached(self):
        with self._publish_static_files("data/cubicweb.css") as req:
            self.assertEqual(200, req.status_out)
            self.assertIn("last-modified", req.headers_out)
        next_headers = {
            "if-modified-since": req.get_response_header("last-modified", raw=True),
        }
        with self._publish_static_files("data/cubicweb.css", next_headers) as req:
            self.assertEqual(304, req.status_out)


class StaticDirectoryControllerCWTC(staticfilespublishermixin, WebCWTC):
    def test_check_static_dir_access(self):
        """write a file in the static directory and test the access"""
        staticdir = osp.join(self.vreg.config.static_directory)
        if not os.path.exists(staticdir):
            os.makedirs(staticdir)
        filename = osp.join(staticdir, "test")
        with open(filename, "a"):
            with self._publish_static_files("static/test") as req:
                self.assertEqual(200, req.status_out)


class DataControllerCWTC(staticfilespublishermixin, WebCWTC):
    tags = WebCWTC.tags | Tags("static_controller", "data", "http")

    def _check_datafile_ok(self, fname):
        with self._publish_static_files(fname) as req:
            self.assertEqual(200, req.status_out)
            self.assertIn("last-modified", req.headers_out)
            self.assertIn("expires", req.headers_out)
            self.assertEqual(
                req.get_response_header("cache-control"), {"max-age": 604800}
            )
        next_headers = {
            "if-modified-since": req.get_response_header("last-modified", raw=True),
        }
        with self._publish_static_files(fname, next_headers) as req:
            self.assertEqual(304, req.status_out)

    def _check_datafile_redirect(self, fname, expected):
        with self._publish_static_files(fname) as req:
            self.assertEqual(302, req.status_out)
            self.assertEqual(
                req.get_response_header("location"), req.base_url() + expected
            )

    def _check_no_datafile(self, fname):
        with self._publish_static_files(fname) as req:
            self.assertEqual(404, req.status_out)

    def test_static_data_mode(self):
        hash = self.vreg.config.instance_md5_version()
        self.assertEqual(32, len(hash))

        with tempattr(self.vreg.config, "mode", "test"):
            self._check_datafile_ok("data/cubicweb.css")
            self._check_no_datafile("data/does/not/exist")
            self._check_no_datafile("data/%s/cubicweb.css" % ("0" * len(hash)))

        with tempattr(self.vreg.config, "mode", "notest"):
            self.config._init_base_url()  # reset config.datadir_url
            self._check_datafile_redirect(
                "data/cubicweb.css", "data/%s/cubicweb.css" % hash
            )
            self._check_datafile_ok("data/%s/cubicweb.css" % hash)
            self._check_no_datafile("data/%s/does/not/exist" % hash)
            self._check_datafile_redirect(
                "data/%s/does/not/exist" % ("0" * len(hash)),
                "data/{}/{}/does/not/exist".format(hash, "0" * len(hash)),
            )


class ConcatFilesCWTC(WebCWTC):
    tags = WebCWTC.tags | Tags("static_controller", "concat")

    def tearDown(self):
        super().tearDown()
        self._cleanup_concat_cache()

    def _cleanup_concat_cache(self):
        uicachedir = osp.join(self.config.apphome, "uicache")
        for fname in glob.glob(osp.join(uicachedir, "cache_concat_*")):
            os.unlink(osp.join(uicachedir, fname))

    @contextmanager
    def _publish_js_files(self, js_files):
        with self.admin_access.web_request() as req:
            head = HTMLHead(req)
            url = head.concat_urls([req.data_url(js_file) for js_file in js_files])[
                len(req.base_url()) :
            ]
            req._url = url
            res = self.app_handle_request(req)
            yield res, req

    def expected_content(self, js_files):
        content = b""
        for js_file in js_files:
            dirpath, rid = self.config.locate_resource(js_file)
            if dirpath is not None:  # ignore resources not found
                with open(osp.join(dirpath, rid), "rb") as f:
                    content += f.read() + b"\n"
        return content

    def test_cache(self):
        js_files = ("cubicweb.ajax.js", "jquery.js")
        with self._publish_js_files(js_files) as (result, req):
            self.assertNotEqual(404, req.status_out)
            # check result content
            self.assertEqual(result, self.expected_content(js_files))
            # make sure we kept a cached version on filesystem
            concat_hander = ConcatFilesHandler(self.config)
            filepath = concat_hander.build_filepath(js_files)
            self.assertTrue(osp.isfile(filepath))

    def test_invalid_file_in_debug_mode(self):
        js_files = ("cubicweb.ajax.js", "dummy.js")
        # in debug mode, an error is raised
        self.config.debugmode = True
        try:
            with self._publish_js_files(js_files) as (result, req):
                # print result
                self.assertEqual(404, req.status_out)
        finally:
            self.config.debugmode = False

    def test_invalid_file_in_production_mode(self):
        js_files = ("cubicweb.ajax.js", "dummy.js")
        with self._publish_js_files(js_files) as (result, req):
            self.assertNotEqual(404, req.status_out)
            # check result content
            self.assertEqual(result, self.expected_content(js_files))


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



from logilab.mtconverter import guess_mimetype_and_encoding
from cubicweb.entities import AnyEntity, fetch_config


class File(AnyEntity):
    """customized class for File entities"""

    __regid__ = "File"
    fetch_attrs, cw_fetch_order = fetch_config(["data_name", "title"])

    def set_format_and_encoding(self):
        """try to set format and encoding according to known values (filename,
        file content, format, encoding).

        This method must be called in a before_[add|update]_entity hook else it
        won't have any effect.
        """
        assert "data" in self.cw_edited, "missing mandatory attribute data"
        if self.cw_edited.get("data"):
            if hasattr(self.data, "filename") and not self.cw_edited.get("data_name"):
                self.cw_edited["data_name"] = self.data.filename
        else:
            self.cw_edited["data_format"] = None
            self.cw_edited["data_encoding"] = None
            self.cw_edited["data_name"] = None
            return
        if "data_format" in self.cw_edited:
            format = self.cw_edited.get("data_format")
        else:
            format = None
        if "data_encoding" in self.cw_edited:
            encoding = self.cw_edited.get("data_encoding")
        else:
            encoding = None
        if not (format and encoding):
            format, encoding = guess_mimetype_and_encoding(
                data=self.cw_edited.get("data"),
                # use get and not get_value since data has changed, we only
                # want to consider explicitly specified values, not old ones
                filename=self.cw_edited.get("data_name"),
                format=format,
                encoding=encoding,
                fallbackencoding=self._cw.encoding,
            )
            if format:
                self.cw_edited["data_format"] = str(format)
            if encoding:
                self.cw_edited["data_encoding"] = str(encoding)


class UnResizeable(Exception):
    pass



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""base classes to handle tabbed views"""


from logilab.common.deprecation import class_renamed
from logilab.mtconverter import xml_escape

from cubicweb import NoSelectableObject, role
from cubicweb import _
from cubicweb import uilib, utils
from cubicweb.predicates import partial_has_related_entities
from cubicweb_web import tags
from cubicweb_web.view import EntityView
from cubicweb_web.views import primary


class LazyViewMixin:
    """provides two convenience methods for the tab machinery.

    Can also be used to lazy-load arbitrary views.
    """

    def _prepare_bindings(self, vid, reloadable):
        self._cw.add_onload(
            """
  jQuery('#lazy-%(vid)s').bind('%(event)s', function(event) {
     loadNow('#lazy-%(vid)s', '#%(vid)s-hole', %(reloadable)s);
  });"""
            % {
                "event": "load_%s" % vid,
                "vid": vid,
                "reloadable": str(reloadable).lower(),
            }
        )

    def lazyview(
        self,
        vid,
        rql=None,
        eid=None,
        rset=None,
        tabid=None,
        reloadable=False,
        show_spinbox=True,
        w=None,
    ):
        """a lazy version of wview"""
        w = w or self.w
        self._cw.add_js("cubicweb.ajax.js")
        # the form is copied into urlparams to please the inner views
        # that might want to take params from it
        # beware of already present rql or eid elements
        # to be safe of collision a proper argument passing protocol
        # (with namespaces) should be used instead of the current
        # ad-hockery
        urlparams = self._cw.form.copy()
        urlparams.pop("rql", None)
        urlparams.pop("eid", None)
        urlparams.update({"vid": vid, "fname": "view"})
        if rql:
            urlparams["rql"] = rql
        elif eid:
            urlparams["eid"] = eid
        elif rset:
            urlparams["rql"] = rset.printable_rql()
        if tabid is None:
            tabid = uilib.domid(vid)
        self.w(
            '<div id="lazy-%s" cubicweb:loadurl="%s">',
            tabid,
            xml_escape(self._cw.build_url("ajax", **urlparams)),
        )
        if show_spinbox:
            # Don't use ``alt`` since image is a *visual* helper for ajax
            self.w(
                '<img style="display: none" src="%s" alt="" id="%s-hole"/>',
                xml_escape(self._cw.data_url("loading.gif")),
                tabid,
            )
        else:
            self.w('<div id="%s-hole"></div>', tabid)
        self.w(
            '<noscript><p>%s <a id="seo-%s" href="%s">%s</a></p></noscript>',
            xml_escape(self._cw._("Link:")),
            tabid,
            xml_escape(self._cw.build_url(**urlparams)),
            xml_escape(self._cw._(tabid)),
        )
        self.w("</div>")
        self._prepare_bindings(tabid, reloadable)

    def forceview(self, vid):
        """trigger an event that will force immediate loading of the view on dom
        readyness
        """
        self._cw.add_onload(uilib.js.triggerLoad(vid))


class TabsMixin(LazyViewMixin):
    """a tab mixin to easily get jQuery based, lazy, ajax tabs"""

    lazy = True

    @property
    def cookie_name(self):
        return str("%s_active_tab" % self._cw.vreg.config.appid)

    def active_tab(self, default):
        if "tab" in self._cw.form:
            return self._cw.form["tab"]
        cookies = self._cw.get_cookie()
        cookiename = self.cookie_name
        activetab = cookies.get(cookiename)
        if activetab is None:
            domid = uilib.domid(default)
            self._cw.set_cookie(cookiename, domid)
            return domid
        return activetab.value

    def prune_tabs(self, tabs, default_tab):
        selected_tabs = []
        may_be_active_tab = self.active_tab(default_tab)
        active_tab = uilib.domid(default_tab)
        viewsvreg = self._cw.vreg["views"]
        for tab in tabs:
            if isinstance(tab, str):
                tabid, tabkwargs = tab, {}
            else:
                tabid, tabkwargs = tab
                tabkwargs = tabkwargs.copy()
            tabkwargs.setdefault("rset", self.cw_rset)
            vid = tabkwargs.get("vid", tabid)
            domid = uilib.domid(tabid)
            try:
                viewsvreg.select(vid, self._cw, tabid=domid, **tabkwargs)
            except NoSelectableObject:
                continue
            selected_tabs.append((tabid, domid, tabkwargs))
            if domid == may_be_active_tab:
                active_tab = domid
        return selected_tabs, active_tab

    def render_tabs(self, tabs, default, entity=None):
        # delegate to the default tab if there is more than one entity
        # in the result set (tabs are pretty useless there)
        if entity and len(self.cw_rset) > 1:
            entity.view(default, w=self.w)
            return
        self._cw.add_css("jquery.ui.css")
        self._cw.add_js(("jquery.ui.js", "cubicweb.ajax.js", "jquery.cookie.js"))
        # prune tabs : not all are to be shown
        tabs, active_tab = self.prune_tabs(tabs, default)
        # build the html structure
        uid = entity and entity.eid or utils.make_uid("tab")
        self.w('<div id="entity-tabs-%s">', uid)
        self.w("<ul>")
        active_tab_idx = None
        for i, (tabid, domid, tabkwargs) in enumerate(tabs):
            self.w("<li>")
            self.w('<a href="#%s">', domid)
            self.w(tabkwargs.pop("label", self._cw._(tabid)))
            self.w("</a>")
            self.w("</li>")
            if domid == active_tab:
                active_tab_idx = i
        self.w("</ul>")
        for tabid, domid, tabkwargs in tabs:
            self.w('<div id="%s">', domid)
            if self.lazy:
                tabkwargs.setdefault("tabid", domid)
                tabkwargs.setdefault("vid", tabid)
                self.lazyview(**tabkwargs)
            else:
                self._cw.view(tabid, w=self.w, **tabkwargs)
            self.w("</div>")
        self.w("</div>")
        # call the setTab() JS function *after* each tab is generated
        # because the callback binding needs to be done before
        # XXX make work history: true
        if self.lazy:
            self._cw.add_onload(
                """
  jQuery('#entity-tabs-%(uid)s').tabs(
    { active: %(tabindex)s,
      activate: function(event, ui) {
        setTab(ui.newPanel.attr('id'), '%(cookiename)s');
      }
    });
  setTab('%(domid)s', '%(cookiename)s');
"""
                % {
                    "tabindex": active_tab_idx,
                    "domid": active_tab,
                    "uid": uid,
                    "cookiename": self.cookie_name,
                }
            )
        else:
            self._cw.add_onload(
                "jQuery('#entity-tabs-%(uid)s').tabs({active: %(tabindex)s});"
                % {"tabindex": active_tab_idx, "uid": uid}
            )


class EntityRelationView(EntityView):
    """view displaying entity related stuff.
    Such a view _must_ provide the rtype, target and vid attributes :

    Example :

    class ProjectScreenshotsView(EntityRelationView):
        '''display project's screenshots'''
        __regid__ = title = _('projectscreenshots')
        __select__ = EntityRelationView.__select__ & is_instance('Project')
        rtype = 'screenshot'
        role = 'subject'
        vid = 'gallery'

    in this example, entities related to project entity by the 'screenshot'
    relation (where the project is subject of the relation) will be displayed
    using the 'gallery' view.
    """

    __select__ = EntityView.__select__ & partial_has_related_entities()
    vid = "list"
    # to be defined in concrete classes
    rtype = title = None

    def cell_call(self, row, col):
        rset = self.cw_rset.get_entity(row, col).related(self.rtype, role(self))
        self.w('<div class="mainInfo">')
        if self.title:
            self.w(tags.h1(self._cw._(self.title)))
        self.wview(self.vid, rset, "noresult")
        self.w("</div>")


class TabbedPrimaryView(TabsMixin, primary.PrimaryView):
    __abstract__ = True  # don't register

    tabs = [_("main_tab")]
    default_tab = "main_tab"

    def render_entity(self, entity):
        self.render_entity_toolbox(entity)
        self.w('<div class="tabbedprimary"></div>')
        self.render_entity_title(entity)
        self.render_tabs(self.tabs, self.default_tab, entity)


TabedPrimaryView = class_renamed("TabedPrimaryView", TabbedPrimaryView)


class PrimaryTab(primary.PrimaryView):
    __regid__ = "main_tab"
    title = None  # should not appear in possible views

    def is_primary(self):
        return True

    def render_entity_title(self, entity):
        pass

    def render_entity_toolbox(self, entity):
        pass



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""vcard import / export"""

from logilab.common.deprecation import class_renamed

from cubicweb import _
from cubicweb.predicates import is_instance
from cubicweb_web.view import EntityView

VCARD_PHONE_TYPES = {"home": "HOME", "office": "WORK", "mobile": "CELL", "fax": "FAX"}


class VCardCWUserView(EntityView):
    """export a person information as a vcard"""

    __regid__ = "vcard"
    title = _("vcard")
    templatable = False
    content_type = "text/x-vcard"
    __select__ = is_instance("CWUser")

    def set_request_content_type(self):
        """overriden to set a .vcf filename"""
        self._cw.set_content_type(self.content_type, filename="vcard.vcf")

    def cell_call(self, row, col):
        self.vcard_header()
        self.vcard_content(self.cw_rset.complete_entity(row, col))
        self.vcard_footer()

    def vcard_header(self):
        self.w("BEGIN:vcard\n")
        self.w("VERSION:3.0\n")

    def vcard_footer(self):
        self.w("NOTE:this card has been generated by CubicWeb\n")
        self.w("END:vcard\n")

    def vcard_content(self, entity):
        who = "{} {}".format(entity.surname or "", entity.firstname or "")
        self.w("FN:%s\n", who)
        self.w("N:%s;;;;\n", entity.login)
        self.w("TITLE:%s\n", who)
        for email in entity.use_email:
            self.w("EMAIL;TYPE=INTERNET:%s\n", email.address)


VCardEuserView = VCardEUserView = class_renamed("VCardEuserView", VCardCWUserView)



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Set of HTML base actions"""


from logilab.common.registry import objectify_predicate, yes

from cubicweb import _
from cubicweb.predicates import (
    EntityPredicate,
    one_line_rset,
    multi_lines_rset,
    one_etype_rset,
    nonempty_rset,
    non_final_entity,
    score_entity,
    authenticated_user,
    match_user_groups,
    match_search_state,
    has_permission,
    has_add_permission,
    is_instance,
)
from cubicweb.schema import display_name
from cubicweb_web import controller, action
from cubicweb_web.views import uicfg, linksearch_select_url, vid_from_rset


class has_editable_relation(EntityPredicate):
    """accept if some relations for an entity found in the result set is
    editable by the logged user.

    See `EntityPredicate` documentation for behaviour when row is not specified.
    """

    def score_entity(self, entity):
        # if user has no update right but it can modify some relation,
        # display action anyway
        form = entity._cw.vreg["forms"].select(
            "edition", entity._cw, entity=entity, mainform=False
        )
        for dummy in form.iter_editable_relations():
            return 1
        for dummy in form.inlined_form_views():
            return 1
        for dummy in form.editable_attributes(strict=True):
            return 1
        return 0


@objectify_predicate
def match_searched_etype(cls, req, rset=None, **kwargs):
    return req.match_search_state(rset)


@objectify_predicate
def view_is_not_default_view(cls, req, rset=None, **kwargs):
    # interesting if it propose another view than the current one
    vid = req.form.get("vid")
    if vid and vid != vid_from_rset(req, rset, req.vreg.schema):
        return 1
    return 0


@objectify_predicate
def addable_etype_empty_rset(cls, req, rset=None, **kwargs):
    if rset is not None and not rset.rowcount:
        rqlst = rset.syntax_tree()
        if len(rqlst.children) > 1:
            return 0
        select = rqlst.children[0]
        if len(select.defined_vars) == 1 and len(select.solutions) == 1:
            rset._searched_etype = next(iter(select.solutions[0].values()))
            eschema = req.vreg.schema.entity_schema_for(rset._searched_etype)
            if not (
                eschema.final or eschema.is_subobject(strict=True)
            ) and eschema.has_perm(req, "add"):
                return 1
    return 0


class has_undoable_transactions(EntityPredicate):
    "Select entities having public (i.e. end-user) undoable transactions."

    def score_entity(self, entity):
        if not entity._cw.vreg.config["undo-enabled"]:
            return 0
        if entity._cw.cnx.undoable_transactions(eid=entity.eid):
            return 1
        else:
            return 0


# generic 'main' actions #######################################################


class SelectAction(action.Action):
    """base class for link search actions. By default apply on
    any size entity result search it the current state is 'linksearch'
    if accept match.
    """

    __regid__ = "select"
    __select__ = (
        match_search_state("linksearch") & nonempty_rset() & match_searched_etype()
    )

    title = _("select")
    category = "mainactions"
    order = 0

    def url(self):
        return linksearch_select_url(self._cw, self.cw_rset)


class CancelSelectAction(action.Action):
    __regid__ = "cancel"
    __select__ = match_search_state("linksearch")

    title = _("cancel select")
    category = "mainactions"
    order = 10

    def url(self):
        target, eid, r_type, searched_type = self._cw.search_state[1]
        return self._cw.build_url(str(eid), vid="edition", __mode="normal")


class ViewAction(action.Action):
    __regid__ = "view"
    __select__ = (
        action.Action.__select__
        & match_user_groups("users", "managers")
        & view_is_not_default_view()
        & non_final_entity()
    )

    title = _("view")
    category = "mainactions"
    order = 0

    def url(self):
        params = self._cw.form.copy()
        for param in ("vid", "__message") + controller.NAV_FORM_PARAMETERS:
            params.pop(param, None)
        if self._cw.ajax_request:
            path = "view"
            if self.cw_rset is not None:
                params = {"rql": self.cw_rset.printable_rql()}
        else:
            path = self._cw.relative_path(includeparams=False)
        return self._cw.build_url(path, **params)


class ModifyAction(action.Action):
    __regid__ = "edit"
    __select__ = action.Action.__select__ & one_line_rset() & has_editable_relation()

    title = _("modify")
    category = "mainactions"
    order = 10

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid="edition")


class MultipleEditAction(action.Action):
    __regid__ = "muledit"  # XXX get strange conflicts if id='edit'
    __select__ = (
        action.Action.__select__
        & multi_lines_rset()
        & one_etype_rset()
        & has_permission("update")
    )

    title = _("modify")
    category = "mainactions"
    order = 10

    def url(self):
        return self._cw.build_url(
            "view", rql=self.cw_rset.printable_rql(), vid="muledit"
        )


# generic "more" actions #######################################################


class ManagePermissionsAction(action.Action):
    __regid__ = "managepermission"
    __select__ = (
        action.Action.__select__
        & one_line_rset()
        & non_final_entity()
        & match_user_groups("managers")
    )

    title = _("manage permissions")
    category = "moreactions"
    order = 15

    def url(self):
        return self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0).absolute_url(
            vid="security"
        )


class DeleteAction(action.Action):
    __regid__ = "delete"
    __select__ = action.Action.__select__ & has_permission("delete")

    title = _("delete")
    category = "moreactions"
    order = 20

    def url(self):
        if len(self.cw_rset) == 1:
            entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
            return self._cw.build_url(entity.rest_path(), vid="deleteconf")
        return self._cw.build_url(rql=self.cw_rset.printable_rql(), vid="deleteconf")


class CopyAction(action.Action):
    __regid__ = "copy"
    __select__ = action.Action.__select__ & one_line_rset() & has_permission("add")

    title = _("copy")
    category = "moreactions"
    order = 30

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid="copy")


class AddNewAction(MultipleEditAction):
    """when we're seeing more than one entity with the same type, propose to
    add a new one
    """

    __regid__ = "addentity"
    __select__ = action.Action.__select__ & (
        addable_etype_empty_rset()
        | (multi_lines_rset() & one_etype_rset() & has_add_permission())
    )

    category = "moreactions"
    order = 40

    @property
    def rsettype(self):
        if self.cw_rset:
            return self.cw_rset.description[0][0]
        return self.cw_rset._searched_etype

    @property
    def title(self):
        return self._cw.__("add a %s" % self.rsettype)  # generated msgid

    def url(self):
        return (
            self._cw.vreg["etypes"].etype_class(self.rsettype).cw_create_url(self._cw)
        )


class AddRelatedActions(action.Action):
    """fill 'addrelated' sub-menu of the actions box"""

    __regid__ = "addrelated"
    __select__ = action.Action.__select__ & one_line_rset() & non_final_entity()

    submenu = _("addrelated")
    order = 17

    def fill_menu(self, box, menu):
        # when there is only one item in the sub-menu, replace the sub-menu by
        # item's title prefixed by 'add'
        menu.label_prefix = self._cw._("add")
        super().fill_menu(box, menu)

    def redirect_params(self, entity):
        return {
            "__redirectpath": entity.rest_path(),  # should not be url quoted!
            "__redirectvid": self._cw.form.get("vid", ""),
        }

    def actual_actions(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        eschema = entity.e_schema
        params = self.redirect_params(entity)
        for rschema, teschema, role in self.add_related_schemas(entity):
            if (
                rschema.role_relation_definition(
                    eschema, teschema, role
                ).role_cardinality(role)
                in "1?"
            ):
                if entity.related(rschema, role):
                    continue
            if role == "subject":
                label = f"add {eschema} {rschema} {teschema} {role}"
                url = self.linkto_url(entity, rschema, teschema, "object", **params)
            else:
                label = f"add {teschema} {rschema} {eschema} {role}"
                url = self.linkto_url(entity, rschema, teschema, "subject", **params)
            yield self.build_action(self._cw._(label), url)

    def add_related_schemas(self, entity):
        """this is actually used ui method to generate 'addrelated' actions from
        the schema.

        If you don't want any auto-generated actions, you should overrides this
        method to return an empty list. If you only want some, you can configure
        them by using uicfg.actionbox_appearsin_addmenu
        """
        appearsin_addmenu = self._cw.vreg["uicfg"].select(
            "actionbox_appearsin_addmenu", self._cw, entity=entity
        )
        req = self._cw
        eschema = entity.e_schema
        for role, rschemas in (
            ("subject", list(eschema.subject_relations.values())),
            ("object", list(eschema.object_relations.values())),
        ):
            for rschema in rschemas:
                if rschema.final:
                    continue
                for teschema in rschema.targets(eschema, role):
                    if not appearsin_addmenu.etype_get(
                        eschema, rschema, role, teschema
                    ):
                        continue
                    rdef = rschema.role_relation_definition(eschema, teschema, role)
                    # RRQLExpression may concern the permissions on the relation
                    # itself or on its subject / object entities. In order to be
                    # able to check the validity of thoses RQL we need to
                    # provide both subject (fromeid) and object (toeid) eids.
                    if not rdef.has_perm(
                        req, "add", toeid=entity.eid, fromeid=entity.eid
                    ):
                        continue
                    # check the target types can be added as well
                    if teschema.may_have_permission("add", req):
                        yield rschema, teschema, role

    def linkto_url(self, entity, rtype, etype, target, **kwargs):
        return (
            self._cw.vreg["etypes"]
            .etype_class(etype)
            .cw_create_url(
                self._cw, __linkto=f"{rtype}:{entity.eid}:{target}", **kwargs
            )
        )


class ViewSameCWEType(action.Action):
    """when displaying the schema of a CWEType, offer to list entities of that type"""

    __regid__ = "entitiesoftype"
    __select__ = (
        one_line_rset() & is_instance("CWEType") & score_entity(lambda x: not x.final)
    )
    category = "mainactions"
    order = 40

    @property
    def etype(self):
        return self.cw_rset.get_entity(0, 0).name

    @property
    def title(self):
        return (
            self._cw.__("view all %s")
            % display_name(self._cw, self.etype, "plural").lower()
        )

    def url(self):
        return self._cw.build_url(self.etype)


# logged user actions #########################################################


class UserPreferencesAction(action.Action):
    __regid__ = "myprefs"
    __select__ = authenticated_user()

    title = _("user preferences")
    category = "useractions"
    order = 10

    def url(self):
        return self._cw.build_url(self.__regid__)


class UserInfoAction(action.Action):
    __regid__ = "myinfos"
    __select__ = authenticated_user()

    title = _("profile")
    category = "useractions"
    order = 20

    def url(self):
        return self._cw.build_url("cwuser/%s" % self._cw.user.login, vid="edition")


class LogoutAction(action.Action):
    __regid__ = "logout"
    __select__ = authenticated_user()

    title = _("logout")
    category = "useractions"
    order = 30

    def url(self):
        return self._cw.build_url(self.__regid__)


# site actions ################################################################


class ManagersAction(action.Action):
    __abstract__ = True
    __select__ = match_user_groups("managers")

    category = "siteactions"

    def url(self):
        return self._cw.build_url(self.__regid__)


class SiteConfigurationAction(ManagersAction):
    __regid__ = "siteconfig"
    title = _("site configuration")
    order = 10
    category = "manage"


class ManageAction(ManagersAction):
    __regid__ = "manage"
    title = _("manage")
    order = 20


# footer actions ###############################################################


class PoweredByAction(action.Action):
    __regid__ = "poweredby"
    __select__ = yes()

    category = "footer"
    order = 3
    title = _("Powered by CubicWeb")

    def url(self):
        return "http://www.cubicweb.org"


# default actions ui configuration ###########################################

addmenu = uicfg.actionbox_appearsin_addmenu
addmenu.tag_object_of(("*", "relation_type", "CWRType"), True)
addmenu.tag_object_of(("*", "from_entity", "CWEType"), False)
addmenu.tag_object_of(("*", "to_entity", "CWEType"), False)
addmenu.tag_object_of(("*", "in_group", "CWGroup"), True)
addmenu.tag_object_of(("*", "bookmarked_by", "CWUser"), True)



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

import copy
import warnings

from yams.buildobjs import RelationDefinition, EntityType

from cubicweb.devtools.testlib import BaseTestCase
from cubicweb.schema import build_schema_from_namespace
from cubicweb_web import uihelper, formwidgets as fwdgs
from cubicweb_web.devtools.testlib import WebCWTC
from cubicweb_web.views import uicfg

abaa = uicfg.actionbox_appearsin_addmenu


class UICFGTC(WebCWTC):
    def test_default_actionbox_appearsin_addmenu_config(self):
        self.assertFalse(abaa.etype_get("TrInfo", "wf_info_for", "object", "CWUser"))


class DefinitionOrderCWTC(WebCWTC):
    """This test check that when multiple definition could match a key, only
    the more accurate apply"""

    def setUp(self):
        super().setUp()
        for rtag in (uicfg.autoform_section, uicfg.autoform_field_kwargs):
            rtag._old_tagdefs = copy.deepcopy(rtag._tagdefs)
        new_def = (
            (("*", "login", "*"), {"formtype": "main", "section": "hidden"}),
            (("*", "login", "*"), {"formtype": "muledit", "section": "hidden"}),
            (("CWUser", "login", "*"), {"formtype": "main", "section": "attributes"}),
            (
                ("CWUser", "login", "*"),
                {"formtype": "muledit", "section": "attributes"},
            ),
            (("CWUser", "login", "String"), {"formtype": "main", "section": "inlined"}),
            (
                ("CWUser", "login", "String"),
                {"formtype": "inlined", "section": "attributes"},
            ),
        )
        for key, kwargs in new_def:
            uicfg.autoform_section.tag_subject_of(key, **kwargs)

    def tearDown(self):
        super().tearDown()
        for rtag in (uicfg.autoform_section, uicfg.autoform_field_kwargs):
            rtag._tagdefs = rtag._old_tagdefs

    def test_definition_order_hidden(self):
        result = uicfg.autoform_section.get("CWUser", "login", "String", "subject")
        expected = {"main_inlined", "muledit_attributes", "inlined_attributes"}
        self.assertSetEqual(result, expected)

    def test_uihelper_set_fields_order(self):
        afk_get = uicfg.autoform_field_kwargs.get
        self.assertEqual(afk_get("CWUser", "firstname", "String", "subject"), {})
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            uihelper.set_fields_order("CWUser", ("login", "firstname", "surname"))
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        self.assertEqual(
            afk_get("CWUser", "firstname", "String", "subject"), {"order": 1}
        )

    def test_uicfg_primaryview_set_fields_order(self):
        pvdc = uicfg.primaryview_display_ctrl
        pvdc.set_fields_order("CWUser", ("login", "firstname", "surname"))
        self.assertEqual(pvdc.get("CWUser", "login", "String", "subject"), {"order": 0})
        self.assertEqual(
            pvdc.get("CWUser", "firstname", "String", "subject"), {"order": 1}
        )
        self.assertEqual(
            pvdc.get("CWUser", "surname", "String", "subject"), {"order": 2}
        )

    def test_uihelper_set_field_kwargs(self):
        afk_get = uicfg.autoform_field_kwargs.get
        self.assertEqual(afk_get("CWUser", "firstname", "String", "subject"), {})
        wdg = fwdgs.TextInput({"size": 30})
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            uihelper.set_field_kwargs("CWUser", "firstname", widget=wdg)
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        self.assertEqual(
            afk_get("CWUser", "firstname", "String", "subject"), {"widget": wdg}
        )

    def test_uihelper_hide_fields(self):
        # original conf : in_group is edited in 'attributes' section everywhere
        section_conf = uicfg.autoform_section.get("CWUser", "in_group", "*", "subject")
        self.assertCountEqual(section_conf, ["main_attributes", "muledit_attributes"])
        # hide field in main form
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            uihelper.hide_fields("CWUser", ("login", "in_group"))
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        section_conf = uicfg.autoform_section.get("CWUser", "in_group", "*", "subject")
        self.assertCountEqual(section_conf, ["main_hidden", "muledit_attributes"])
        # hide field in muledit form
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            uihelper.hide_fields("CWUser", ("login", "in_group"), formtype="muledit")
            self.assertEqual(len(w), 1)
            self.assertTrue(issubclass(w[-1].category, DeprecationWarning))
        section_conf = uicfg.autoform_section.get("CWUser", "in_group", "*", "subject")
        self.assertCountEqual(section_conf, ["main_hidden", "muledit_hidden"])

    def test_uihelper_formconfig(self):
        afk_get = uicfg.autoform_field_kwargs.get

        class CWUserFormConfig(uihelper.FormConfig):
            etype = "CWUser"
            hidden = ("in_group",)
            fields_order = ("login", "firstname")

        section_conf = uicfg.autoform_section.get("CWUser", "in_group", "*", "subject")
        self.assertCountEqual(section_conf, ["main_hidden", "muledit_attributes"])
        self.assertEqual(
            afk_get("CWUser", "firstname", "String", "subject"), {"order": 1}
        )


class UicfgRegistryCWTC(WebCWTC):
    def test_default_uicfg_object(self):
        "CW default ui config objects must be registered in uicfg registry"
        onames = ("autoform_field", "autoform_section", "autoform_field_kwargs")
        for oname in onames:
            obj = self.vreg["uicfg"].select_or_none(oname)
            self.assertIsNotNone(obj, "%s not found in uicfg registry" % oname)

    def test_custom_uicfg(self):
        ASRT = uicfg.AutoformSectionRelationTags
        custom_afs = ASRT()
        custom_afs.__select__ = ASRT.__select__ & ASRT.__select__
        self.vreg["uicfg"].register(custom_afs)
        obj = self.vreg["uicfg"].select_or_none("autoform_section")
        self.assertIs(obj, custom_afs)


def _schema():
    class Personne(EntityType):
        pass

    class Societe(EntityType):
        pass

    class Tag(EntityType):
        pass

    class travaille(RelationDefinition):
        subject = "Personne"
        object = "Societe"

    class tags(RelationDefinition):
        subject = "Tag"
        object = ("Personne", "Societe", "Tag")

    return build_schema_from_namespace(locals().items())


class AutoformSectionTC(BaseTestCase):
    def test_derivation(self):
        schema = _schema()
        afs = uicfg.AutoformSectionRelationTags()
        afs.tag_subject_of(("Personne", "travaille", "*"), "main", "relations")
        afs.tag_object_of(("*", "travaille", "Societe"), "main", "relations")
        afs.tag_subject_of(("Tag", "tags", "*"), "main", "relations")

        afs2 = afs.derive(__name__, afs.__select__)
        afs2.tag_subject_of(("Personne", "travaille", "*"), "main", "attributes")
        afs2.tag_object_of(("*", "travaille", "Societe"), "main", "attributes")
        afs2.tag_subject_of(("Tag", "tags", "Societe"), "main", "attributes")

        afs.init(schema)
        afs2.init(schema)

        self.assertEqual(
            afs2.etype_get("Tag", "tags", "subject", "Personne"),
            {"main_relations", "muledit_hidden", "inlined_relations"},
        )
        self.assertEqual(
            afs2.etype_get("Tag", "tags", "subject", "Societe"),
            {"main_attributes", "muledit_hidden", "inlined_attributes"},
        )
        self.assertEqual(
            afs2.etype_get("Personne", "travaille", "subject", "Societe"),
            {"main_attributes", "muledit_hidden", "inlined_attributes"},
        )
        self.assertEqual(
            afs2.etype_get("Societe", "travaille", "object", "Personne"),
            {"main_attributes", "muledit_hidden", "inlined_attributes"},
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
from cubicweb.devtools import BASE_URL

from cubicweb_web.devtools.testlib import WebCWTC


class BreadCrumbsCWTC(WebCWTC):
    def test_base(self):
        with self.admin_access.web_request() as req:
            f1 = req.create_entity("Folder", name="par&ent")
            f2 = req.create_entity("Folder", name="chi&ld")
            req.cnx.execute(
                "SET F2 filed_under F1 WHERE F1 eid %(f1)s, F2 eid %(f2)s",
                {"f1": f1.eid, "f2": f2.eid},
            )
            req.cnx.commit()
            self.assertEqual(
                f2.view("breadcrumbs"),
                f'<a href="{BASE_URL}folder/{f2.eid}" title="">' f"chi&amp;ld</a>",
            )
            childrset = f2.as_rset()
            ibc = self.vreg["ctxcomponents"].select("breadcrumbs", req, rset=childrset)
            f_list = []
            ibc.render(f_list.append)
            self.assertMultiLineEqual(
                f'<span id="breadcrumbs" class="pathbar">&#160;&gt;&#160;'
                f'<a href="{BASE_URL}Folder">Folder_plural</a>'
                f'&#160;&gt;&#160;<a href="{BASE_URL}folder/{f1.eid}" '
                f'title="">par&amp;ent</a>&#160;&gt;&#160;\n'
                f'<a href="{BASE_URL}folder/{f2.eid}" title="">'
                f"chi&amp;ld</a></span>",
                "".join(f_list),
            )


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2023-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common.registry import Predicate


class paginated_rset(Predicate):
    """Return 1 or more for result set with more rows than one or more page
    size.  You can specify expected number of pages to the initializer (default
    to one), and you'll get that number of pages as score if the result set is
    big enough.

    Page size is searched in (respecting order):
    * a `page_size` argument
    * a `page_size` form parameters
    * the `navigation.page-size` property (see :ref:`PersistentProperties`)
    """

    def __init__(self, nbpages=1):
        assert nbpages > 0
        self.nbpages = nbpages

    def __call__(self, cls, req, rset=None, **kwargs):
        if rset is None:
            return 0
        page_size = kwargs.get("page_size")
        if page_size is None:
            page_size = req.form.get("page_size")

            if page_size is not None:
                if page_size.isdecimal():
                    page_size = int(page_size)
                else:
                    page_size = None

            if page_size is None:
                page_size_prop = getattr(
                    cls, "page_size_property", "navigation.page-size"
                )
                page_size = req.property_value(page_size_prop)
        if len(rset) <= (page_size * self.nbpages):
            return 0
        return self.nbpages



# copyright 2003-2024 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""inline help system, rendering ReST files in the `wdoc` subdirectory of
CubicWeb and cubes

"""


from itertools import chain
from os.path import join
from xml.etree.ElementTree import parse

from logilab.common.registry import yes

from cubicweb import _
from cubicweb.predicates import match_form_params
from cubicweb.uilib import rest_publish
from cubicweb_web.view import StartupView
from cubicweb_web import NotFound, action


# table of content management #################################################


def build_toc_index(node, parent, index):
    try:
        nodeidx = node.attrib["resource"]
        assert nodeidx not in index, nodeidx
        index[nodeidx] = node, parent
    except KeyError:
        pass
    for child in node:
        build_toc_index(child, node, index)


def get_insertion_point(section, index):
    if section.attrib.get("insertafter"):
        snode, node = index[section.attrib["insertafter"]]
        idx = list(node).index(snode) + 1
    elif section.attrib.get("insertbefore"):
        snode, node = index[section.attrib["insertbefore"]]
        idx = node.getchildren().index(snode)
    elif "appendto" in section.attrib:
        node, _ = index[section.attrib["appendto"]]
        idx = None
    else:
        node, idx = None, None
    return node, idx


def build_toc(config):
    alltocfiles = reversed(tuple(config.locate_all_files("toc.xml")))
    maintoc = parse(next(alltocfiles)).getroot()
    index = {}
    build_toc_index(maintoc, None, index)
    # insert component documentation into the tree according to their toc.xml
    # file
    for fpath in alltocfiles:
        toc = parse(fpath).getroot()
        for section in toc:
            node, idx = get_insertion_point(section, index)
            if node is None:
                continue
            if idx is None:
                node.append(section)
            else:
                node.insert(idx, section)
            build_toc_index(section, node, index)
    return index


def title_for_lang(node, lang):
    fallback_title = None
    for title in node.findall("title"):
        title_lang = title.attrib["{http://www.w3.org/XML/1998/namespace}lang"]
        if title_lang == lang:
            return title.text
        if title_lang == "en":
            fallback_title = title.text
    return fallback_title


def subsections(node):
    return [child for child in node if child.tag == "section"]


# help views ##################################################################


class InlineHelpView(StartupView):
    __select__ = match_form_params("fid")
    __regid__ = "wdoc"
    title = _("site documentation")

    def call(self):
        fid = self._cw.form["fid"]
        vreg = self._cw.vreg
        for lang in chain(
            (self._cw.lang, vreg.property_value("ui.language")),
            vreg.config.available_languages(),
        ):
            rid = f"{fid}_{lang}.rst"
            resourcedir = vreg.config.locate_doc_file(rid)
            if resourcedir:
                break
        else:
            raise NotFound
        self.tocindex = build_toc(vreg.config)
        try:
            node, parent = self.tocindex[fid]
        except KeyError:
            node, parent = None, None
        else:
            self.navigation_links(node, parent)
            self.w('<div class="hr"></div>')
            self.w("<h1>%s</h1>", title_for_lang(node, self._cw.lang))
        with open(join(resourcedir, rid)) as f:
            self.w(rest_publish(self, f.read()))
        if node is not None:
            self.subsections_links(node)
            self.w('<div class="hr"></div>')
            self.navigation_links(node, parent)

    def navigation_links(self, node, parent):
        if parent is None:
            return
        brothers = subsections(parent)
        self.w('<div class="docnav">\n')
        previousidx = brothers.index(node) - 1
        if previousidx >= 0:
            self.navsection(brothers[previousidx], "prev")
        self.navsection(parent, "up")
        nextidx = brothers.index(node) + 1
        if nextidx < len(brothers):
            self.navsection(brothers[nextidx], "next")
        self.w("</div>\n")

    navinfo = {
        "prev": ("", "data/previous.png", _("i18nprevnext_previous")),
        "next": ("", "data/next.png", _("i18nprevnext_next")),
        "up": ("", "data/up.png", _("i18nprevnext_up")),
    }

    def navsection(self, node, navtype):
        htmlclass, imgpath, msgid = self.navinfo[navtype]
        self.w('<span class="%s">', htmlclass)
        self.w("%s : ", self._cw._(msgid))
        self.w(
            '<a href="%s">%s</a>',
            self._cw.build_url("doc/" + node.attrib["resource"]),
            title_for_lang(node, self._cw.lang),
        )
        self.w("</span>\n")

    def subsections_links(self, node, first=True):
        sub = subsections(node)
        if not sub:
            return
        if first:
            self.w('<div class="hr"></div>')
        self.w('<ul class="docsum">')
        for child in sub:
            self.w(
                '<li><a href="%s">%s</a>',
                self._cw.build_url("doc/" + child.attrib["resource"]),
                title_for_lang(child, self._cw.lang),
            )
            self.subsections_links(child, False)
            self.w("</li>")
        self.w("</ul>\n")


class InlineHelpImageView(StartupView):
    __regid__ = "wdocimages"
    __select__ = match_form_params("fid")
    binary = True
    templatable = False
    content_type = "image/png"

    def call(self):
        fid = self._cw.form["fid"]
        for lang in chain(
            (self._cw.lang, self._cw.vreg.property_value("ui.language")),
            self._cw.vreg.config.available_languages(),
        ):
            rid = join("images", f"{fid}_{lang}.png")
            resourcedir = self._cw.vreg.config.locate_doc_file(rid)
            if resourcedir:
                break
        else:
            raise NotFound
        with open(join(resourcedir, rid)) as f:
            self.w(f.read())


class HelpAction(action.Action):
    __regid__ = "help"
    __select__ = yes()

    category = "footer"
    order = 0
    title = _("Help")

    def url(self):
        return self._cw.build_url("doc/main")


class AboutAction(action.Action):
    __regid__ = "about"
    __select__ = yes()

    category = "footer"
    order = 2
    title = _("About this site")

    def url(self):
        return self._cw.build_url("doc/about")



import os

from cubicweb.server import hook
from cubicweb.predicates import is_instance
from cubicweb.entities import adapters

from cubicweb_file.entities import UnResizeable


class UpdateFileHook(hook.Hook):
    """a file has been updated, check data_format/data_encoding consistency"""

    __regid__ = "updatefilehook"
    __select__ = hook.Hook.__select__ & is_instance("File")
    events = (
        "before_add_entity",
        "before_update_entity",
    )
    order = -1  # should be run before other hooks
    category = "hash"

    def __call__(self):
        edited = self.entity.cw_edited
        if "data" in edited:
            self.entity.set_format_and_encoding()
            maxsize = None
            if maxsize and self.entity.data_format.startswith("image/"):
                iimage = self.entity.cw_adapt_to("IImage")
                try:
                    edited["data"] = iimage.resize(maxsize)
                except UnResizeable:
                    # if the resize fails for some reason, do nothing
                    # (original image will be stored)
                    pass

            # thumbnail cache invalidation
            if "update" in self.event and "data" in edited:
                thumb = self.entity.cw_adapt_to("IThumbnail")
                if not thumb:
                    return
                thumbpath = thumb.thumbnail_path()
                if thumbpath:
                    try:
                        os.unlink(thumbpath)
                    except Exception as exc:
                        self.warning(
                            "could not invalidate thumbnail file `%s` " "(cause: %s)",
                            thumbpath,
                            exc,
                        )


class FileIDownloadableAdapter(adapters.IDownloadableAdapter):
    __select__ = is_instance("File")

    # IDownloadable
    def download_url(self, **kwargs):
        # include filename in download url for nicer url
        name = self._cw.url_quote(self.download_file_name())
        path = f"{self.entity.rest_path()}/raw/{name}"
        return self._cw.build_url(path, **kwargs)

    def download_content_type(self):
        return self.entity.data_format

    def download_encoding(self):
        return self.entity.data_encoding

    def download_file_name(self):
        return self.entity.data_name

    def download_data(self):
        return self.entity.data.getvalue()
